from __future__ import print_function as _
from pysb.simulator.base import Simulator, SimulatorException, SimulationResult
import pysb
import pysb.bng
import numpy as np
from scipy.constants import N_A
import os
import re
import subprocess
import tempfile
import time
import logging
from pysb.logging import EXTENDED_DEBUG
import shutil
from pysb.pathfinder import get_path
import sympy

try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import pycuda.driver as cuda
except ImportError:
    cuda = None


class CupSodaSimulator(Simulator):
    """An interface for running cupSODA, a CUDA implementation of LSODA.

    cupSODA is a graphics processing unit (GPU)-based implementation of the
    LSODA simulation algorithm (see references). It requires an NVIDIA GPU
    card with support for the CUDA framework version 7 or above. Further
    details of cupSODA and software can be found on github:
    https://github.com/aresio/cupSODA

    The simplest way to install cupSODA is to use a pre-compiled version,
    which can be downloaded from here:
    https://github.com/aresio/cupSODA/releases

    Parameters
    ----------

    model : pysb.Model
        Model to integrate.
    tspan : vector-like, optional
        Time values at which the integrations are sampled. The first and last
        values define the time range.
    initials : list-like, optional
        Initial species concentrations for all simulations. Dimensions are 
        N_SIMS x number of species.
    param_values : list-like, optional
        Parameters for all simulations. Dimensions are N_SIMS x number of 
        parameters.
    verbose : bool or int, optional
        Verbosity level, see :class:`pysb.simulator.base.Simulator` for
        further details.
    **kwargs: dict, optional
        Extra keyword arguments, including:

        * ``gpu``: Index of GPU to run on (default: 0)
        * ``vol``: System volume; required if model encoded in extrinsic 
          (number) units (default: None)
        * ``obs_species_only``: Only output species contained in observables
          (default: True) 
        * ``cleanup``: Delete all temporary files after the simulation is 
          finished. Includes both BioNetGen and cupSODA files. Useful for 
          debugging (default: True)
        * ``prefix``: Prefix for the temporary directory containing cupSODA
          input and output files (default: model name)
        * ``base_dir``: Directory in which temporary directory with cupSODA
          input and output files are placed (default: system directory
          determined by `tempfile.mkdtemp`)
        * ``integrator``: Name of the integrator to use; see 
          `default_integrator_options` (default: 'cupsoda')
        * ``integrator_options``: A dictionary of keyword arguments to
          supply to the integrator; see `default_integrator_options`.

    Attributes
    ----------

    model : pysb.Model
        Model passed to the constructor.
    tspan : numpy.ndarray
        Time values passed to the constructor.
    initials : numpy.ndarray
        Initial species concentrations for all simulations. Dimensions are 
        number of simulations x number of species.
    param_values : numpy.ndarray
        Parameters for all simulations. Dimensions are number of simulations 
        x number of parameters.
    verbose: bool or int
        Verbosity setting. See the base class
        :class:`pysb.simulator.base.Simulator` for further details.
    gpu : int
        Index of GPU being run on
    vol : float or None
        System volume
    n_blocks: int
        Number of GPU blocks used by the simulator.
    outdir : str
        Directory where cupSODA output files are placed. Input files are
        also placed here.
    opts: dict
        Dictionary of options for the integrator in use.
    integrator : str
        Name of the integrator in use.
    default_integrator_options : dict
        Nested dictionary of default options for all supported integrators.

    Notes
    -----

    1. If `vol` is defined, species amounts and rate constants are assumed
       to be in number units and are automatically converted to concentration
       units before generating the cupSODA input files. The species
       concentrations returned by cupSODA are converted back to number units
       during loading.

    2. If `obs_species_only` is True, only the species contained within 
       observables are output by cupSODA. All other concentrations are set 
       to 'nan'.

    References
    ----------

    1. Nobile M.S., Cazzaniga P., Besozzi D., Mauri G., 2014. GPU-accelerated
       simulations of mass-action kinetics models with cupSODA, Journal of
       Supercomputing, 69(1), pp.17-24.
    2. Petzold, L., 1983. Automatic selection of methods for solving stiff and
       nonstiff systems of ordinary differential equations. SIAM journal on
       scientific and statistical computing, 4(1), pp.136-148.

    """

    _supports = {'multi_initials': True, 'multi_param_values': True}

    _memory_options = {'global': '0', 'shared': '1', 'sharedconstant': '2'}

    default_integrator_options = {
        # some sane default options for a few well-known integrators
        'cupsoda': {
            'max_steps': 20000, # max # of internal iterations (LSODA's MXSTEP)
            'atol': 1e-8,  # absolute tolerance
            'rtol': 1e-8,  # relative tolerance
            'n_blocks': None,  # number of GPU blocks
            'memory_usage': 'sharedconstant'}}  # see _memory_options dict

    def __init__(self, model, tspan=None, initials=None, param_values=None,
                 verbose=False, **kwargs):
        super(CupSodaSimulator, self).__init__(model, tspan=tspan,
                                               initials=initials,
                                               param_values=param_values,
                                               verbose=verbose, **kwargs)
        self.gpu = kwargs.get('gpu', 0)
        self._obs_species_only = kwargs.get('obs_species_only', True)
        self._cleanup = kwargs.get('cleanup', True)
        self._prefix = kwargs.get('prefix', self._model.name.replace('.', '_'))
        self._base_dir = kwargs.get('base_dir', None)
        self.integrator = kwargs.get('integrator', 'cupsoda')

        # generate the equations for the model
        pysb.bng.generate_equations(self._model, self._cleanup, self.verbose)

        # build integrator options list from our defaults and any kwargs
        # passed to this function
        options = {}
        if self.default_integrator_options.get(self.integrator):
            options.update(self.default_integrator_options[
                self.integrator])  # default options
        else:
            raise SimulatorException(
                "Integrator type '" + self.integrator + "' not recognized.")
        options.update(kwargs.get('integrator_options', {}))  # overwrite

        # defaults
        self.opts = options
        self._out_species = None

        # private variables (to reduce the number of function calls)
        self._len_rxns = len(self._model.reactions)
        self._len_species = len(self._model.species)
        self._len_params = len(self._model.parameters)
        self._model_parameters_rules = self._model.parameters_rules()

        # Set cupsoda verbosity level
        logger_level = self._logger.logger.getEffectiveLevel()
        if logger_level <= EXTENDED_DEBUG:
            self._cupsoda_verbose = 2
        elif logger_level <= logging.DEBUG:
            self._cupsoda_verbose = 1
        else:
            self._cupsoda_verbose = 0

        # regex for extracting cupSODA reported running time
        self._running_time_regex = re.compile(r'Running time:\s+(\d+\.\d+)')

    def run(self, tspan=None, initials=None, param_values=None):
        """Perform a set of integrations.

        Returns a :class:`.SimulationResult` object.

        Parameters
        ----------
        tspan : list-like, optional
            Time values at which the integrations are sampled. The first and
            last values define the time range.
        initials : list-like, optional
            Initial species concentrations for all simulations. Dimensions are
            number of simulation x number of species.    
        param_values : list-like, optional
            Parameters for all simulations. Dimensions are number of
            simulations x number of parameters.

        Returns
        -------
        A :class:`SimulationResult` object

        Notes
        -----
        1. An exception is thrown if `tspan` is not defined in either
           `__init__`or `run`.
           
        2. If neither `initials` nor `param_values` are defined in either 
           `__init__` or `run` a single simulation is run with the initial 
           concentrations and parameter values defined in the model.

        """
        super(CupSodaSimulator, self).run(tspan=tspan, initials=initials,
                                          param_values=param_values,
                                          _run_kwargs=[])

        # Create directories for cupSODA input and output files
        self.outdir = tempfile.mkdtemp(prefix=self._prefix + '_',
                                       dir=self._base_dir)

        self._logger.debug("Output directory is %s" % self.outdir)
        _cupsoda_infiles_dir = os.path.join(self.outdir, "INPUT")
        os.mkdir(_cupsoda_infiles_dir)
        self._cupsoda_outfiles_dir = os.path.join(self.outdir, "OUTPUT")

        # Path to cupSODA executable
        bin_path = get_path('cupsoda')

        # Create cupSODA input files
        self._create_input_files(_cupsoda_infiles_dir)

        # Build command
        # ./cupSODA input_model_folder blocks output_folder simulation_
        # file_prefix gpu_number fitness_calculation memory_use dump
        command = [bin_path, _cupsoda_infiles_dir, str(self.n_blocks),
                   self._cupsoda_outfiles_dir, self._prefix, str(self.gpu),
                   '0', self._memory_usage, str(self._cupsoda_verbose)]

        self._logger.info("Running cupSODA: " + ' '.join(command))
        start_time = time.time()
        # Run simulation and return trajectories
        p = subprocess.Popen(command, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        # for line in iter(p.stdout.readline, b''):
        #     if 'Running time' in line:
        #         self._logger.info(line[:-1])
        (p_out, p_err) = p.communicate()
        p_out = p_out.decode('utf-8')
        p_err = p_err.decode('utf-8')
        logger_level = self._logger.logger.getEffectiveLevel()
        if logger_level <= logging.INFO:
            run_time_match = self._running_time_regex.search(p_out)
            if run_time_match:
                self._logger.info('cupSODA reported time: {} '
                                  'seconds'.format(run_time_match.group(1)))
        self._logger.debug('cupSODA stdout:\n' + p_out)
        if p_err:
            self._logger.error('cupsoda strerr:\n' + p_err)
        if p.returncode:
            raise SimulatorException(
                p_out.rstrip("at line") + "\n" + p_err.rstrip())
        tout, trajectories = self._load_trajectories(
            self._cupsoda_outfiles_dir)
        if self._cleanup:
            shutil.rmtree(self.outdir)
        end_time = time.time()
        self._logger.info("cupSODA + I/O time: {} seconds".format(end_time -
                                                                  start_time))
        return SimulationResult(self, tout, trajectories)

    @property
    def _memory_usage(self):
        try:
            return self._memory_options[self.opts['memory_usage']]
        except KeyError:
            raise Exception('memory_usage must be one of %s',
                            self._memory_options.keys())

    @property
    def vol(self):
        vol = self.opts.get('vol', None)
        return vol

    @vol.setter
    def vol(self, volume):
        self.opts['vol'] = volume

    @property
    def n_blocks(self):
        n_blocks = self.opts.get('n_blocks')
        if n_blocks is None:
            default_threads_per_block = 32
            bytes_per_float = 4
            memory_per_thread = (self._len_species + 1) * bytes_per_float
            if cuda is None:
                threads_per_block = default_threads_per_block
            else:
                cuda.init()
                device = cuda.Device(self.gpu)
                attrs = device.get_attributes()
                shared_memory_per_block = attrs[
                    cuda.device_attribute.MAX_SHARED_MEMORY_PER_BLOCK]
                upper_limit_threads_per_block = attrs[
                    cuda.device_attribute.MAX_THREADS_PER_BLOCK]
                max_threads_per_block = min(
                    shared_memory_per_block / memory_per_thread,
                    upper_limit_threads_per_block)
                threads_per_block = min(max_threads_per_block,
                                        default_threads_per_block)
            n_blocks = int(
                np.ceil(1. * len(self.param_values) / threads_per_block))
            self._logger.debug('n_blocks set to {} (used pycuda: {})'.format(
                n_blocks, cuda is not None
            ))
        self.n_blocks = n_blocks
        return n_blocks

    @n_blocks.setter
    def n_blocks(self, n_blocks):
        if not isinstance(n_blocks, int):
            raise ValueError("n_blocks must be an integer")
        if n_blocks <= 0:
            raise ValueError("n_blocks must be greater than 0")
        self.opts['n_blocks'] = n_blocks

    def _create_input_files(self, directory):

        n_sims = len(self.param_values)

        # atol_vector
        with open(os.path.join(directory, "atol_vector"), 'w') as atol_vector:
            for i in range(self._len_species):
                atol_vector.write(str(self.opts.get('atol')))
                if i < self._len_species - 1:
                    atol_vector.write("\n")

        # c_matrix
        with open(os.path.join(directory, "c_matrix"), 'w') as c_matrix:
            cmtx = self._get_cmatrix()
            for i in range(n_sims):
                line = ""
                for j in range(self._len_rxns):
                    if j > 0:
                        line += "\t"
                    line += str(cmtx[i][j])
                c_matrix.write(line)
                if i < n_sims - 1:
                    c_matrix.write("\n")

        # cs_vector
        with open(os.path.join(directory, "cs_vector"), 'w') as cs_vector:
            self._out_species = range(self._len_species)  # species to output
            if self._obs_species_only:
                self._out_species = [False for sp in self._model.species]
                for obs in self._model.observables:
                    for i in obs.species:
                        self._out_species[i] = True
                self._out_species = [i for i in range(self._len_species) if
                                     self._out_species[i] is True]
            for i in range(len(self._out_species)):
                if i > 0:
                    cs_vector.write("\n")
                cs_vector.write(str(self._out_species[i]))

        # left_side
        with open(os.path.join(directory, "left_side"), 'w') as left_side:
            for i in range(self._len_rxns):
                line = ""
                for j in range(self._len_species):
                    if j > 0:
                        line += "\t"
                    stoich = 0
                    for k in self._model.reactions[i]['reactants']:
                        if j == k:
                            stoich += 1
                    line += str(stoich)
                if i < self._len_rxns - 1:
                    left_side.write(line + "\n")
                else:
                    left_side.write(line)

        # max_steps
        with open(os.path.join(directory, "max_steps"), 'w') as mxsteps:
            mxsteps.write(str(self.opts['max_steps']))

        # model_kind
        with open(os.path.join(directory, "modelkind"), 'w') as model_kind:
            # always set modelkind to 'deterministic'
            model_kind.write("deterministic")

        # MX_0
        with open(os.path.join(directory, "MX_0"), 'w') as MX_0:
            mx0 = self.initials
            # if a volume has been defined, rescale populations
            # by N_A*vol to get concentration
            # (NOTE: act on a copy of self.initials, not
            # the original, which we don't want to modify)
            if self.vol:
                mx0 = mx0.copy()
                mx0 /= (N_A * self.vol)
            for i in range(n_sims):
                line = ""
                for j in range(self._len_species):
                    if j > 0:
                        line += "\t"
                    line += str(mx0[i][j])
                MX_0.write(line)
                if i < n_sims - 1:
                    MX_0.write("\n")

        # right_side
        with open(os.path.join(directory, "right_side"), 'w') as right_side:
            for i in range(self._len_rxns):
                line = ""
                for j in range(self._len_species):
                    if j > 0:
                        line += "\t"
                    stochiometry = 0
                    for k in self._model.reactions[i]['products']:
                        if j == k:
                            stochiometry += 1
                    line += str(stochiometry)
                if i < self._len_rxns - 1:
                    right_side.write(line + "\n")
                else:
                    right_side.write(line)

        # rtol
        with open(os.path.join(directory, "rtol"), 'w') as rtol:
            rtol.write(str(self.opts.get('rtol')))

        # t_vector
        with open(os.path.join(directory, "t_vector"), 'w') as t_vector:
            for t in self.tspan:
                t_vector.write(str(float(t)) + "\n")

        # time_max
        with open(os.path.join(directory, "time_max"), 'w') as time_max:
            time_max.write(str(float(self.tspan[-1])))

    def _get_cmatrix(self):
        self._logger.debug("Constructing the c_matrix:")
        c_matrix = np.zeros((len(self.param_values), self._len_rxns))
        par_names = [p.name for p in self._model_parameters_rules]
        rate_mask = np.array([p in self._model_parameters_rules for p in
                              self._model.parameters])
        rate_args = []
        par_vals = self.param_values[:, rate_mask]
        rate_order = []
        for rxn in self._model.reactions:
            rate_args.append([arg for arg in rxn['rate'].atoms(sympy.Symbol) if
                              not arg.name.startswith('__s')])
            reactants = len(rxn['reactants'])
            rate_order.append(reactants)
        output = 0.01 * len(par_vals)
        output = int(output) if output > 1 else 1
        for i in range(len(par_vals)):
            if i % output == 0:
                self._logger.debug(str(int(round(100. * i / len(par_vals)))) +
                                   "%")
            for j in range(self._len_rxns):
                rate = 1.0
                for r in rate_args[j]:
                    if isinstance(r, pysb.Parameter):
                        rate *= par_vals[i][par_names.index(r.name)]
                    elif isinstance(r, pysb.Expression):
                        raise ValueError('cupSODA does not currently support '
                                         'models with Expressions')
                    else:
                        rate *= r
                # volume correction
                if self.vol:
                    rate *= (N_A * self.vol) ** (rate_order[j] - 1)
                c_matrix[i][j] = rate
        self._logger.debug("100%")
        return c_matrix

    def _load_trajectories(self, directory):
        """Read simulation results from output files.

        Returns `tout` and `trajectories` arrays.
        """
        files = [filename for filename in os.listdir(directory) if
                 re.match(self._prefix, filename)]
        if len(files) == 0:
            raise SimulatorException(
                "Cannot find any output files to load data from.")
        if len(files) != len(self.param_values):
            raise SimulatorException(
                "Number of input files (%d) does not match number "
                "of requested simulations (%d)." % (
                len(files), len(self.param_values)))
        n_sims = len(files)
        trajectories = [None] * n_sims
        tout = [None] * n_sims
        traj_n = np.ones((len(self.tspan), self._len_species)) * float('nan')
        tout_n = np.ones(len(self.tspan)) * float('nan')
        # load the data
        indir_prefix = os.path.join(directory, self._prefix)
        for n in range(n_sims):
            trajectories[n] = traj_n.copy()
            tout[n] = tout_n.copy()
            filename = indir_prefix + "_" + str(n)
            if not os.path.isfile(filename):
                raise Exception("Cannot find input file " + filename)
            # determine optimal loading method
            if n == 0:
                (data, use_pandas) = self._test_pandas(filename)
            # load data
            else:
                if use_pandas:
                    data = self._load_with_pandas(filename)
                else:
                    data = self._load_with_openfile(filename)
            # store data
            tout[n] = data[:, 0]
            trajectories[n][:, self._out_species] = data[:, 1:]
            # volume correction
            if self.vol:
                trajectories[n][:, self._out_species] *= (N_A * self.vol)
        return np.array(tout), np.array(trajectories)

    def _test_pandas(self, filename):
        """ calculates the fastest method to load in data

        Parameters
        ----------
        filename : str
            filename to laod in

        Returns
        -------
        np.array, bool

        """
        # using open(filename,...)
        start = time.time()
        data = self._load_with_openfile(filename)
        end = time.time()
        load_time_openfile = end - start

        # using pandas
        if pd:
            start = time.time()
            self._load_with_pandas(filename)
            end = time.time()
            load_time_pandas = end - start
            if load_time_pandas < load_time_openfile:
                return data, True

        return data, False

    @staticmethod
    def _load_with_pandas(filename):
        data = pd.read_csv(filename, sep='\t', skiprows=None,
                           header=None).as_matrix()
        return data

    @staticmethod
    def _load_with_openfile(filename):
        with open(filename, 'r') as f:
            data = [line.rstrip('\n').split() for line in f]
        data = np.array(data, dtype=np.float, copy=False)
        return data


def run_cupsoda(model, tspan, initials=None, param_values=None,
                integrator='cupsoda', cleanup=True, verbose=False, **kwargs):
    """Wrapper method for running cupSODA simulations.
    
    Parameters
    ----------
    See ``CupSodaSimulator`` constructor.
    
    Returns
    -------
    SimulationResult.all : list of record arrays
        List of trajectory sets. The first dimension contains species,
        observables and expressions (in that order)
    """
    sim = CupSodaSimulator(model, tspan=tspan, integrator=integrator,
                           cleanup=cleanup, verbose=verbose, **kwargs)
    simres = sim.run(initials=initials, param_values=param_values)
    return simres.all



# import the pysb module and all its methods and functions
from pysb import *


def catalyze(enz, sub, site, state1, state2, kf, kr, kc):   # function call
    """2-step catalytic process"""                          # reaction name
    r1_name = '%s_assoc_%s' % (enz.name, sub.name)           # name of association reaction for rule
    r2_name = '%s_diss_%s' % (enz.name, sub.name)           # name of dissociation reaction for rule
    E = enz(b=None)                                         # define enzyme state in function
    S = sub({'b': None, site: state1})                      # define substrate state in function
    ES = enz(b=1) % sub({'b': 1, site: state1})             # define state of enzyme:substrate complex
    P = sub({'b': None, site: state2})                      # define state of product
    Rule(r1_name, E + S | ES, kf, kr)                      # rule for enzyme + substrate association (bidirectional)
    Rule(r2_name, ES >> E + P, kc)                          # rule for enzyme:substrate dissociation  (unidirectional)
   
# instantiate a model
Model()

# declare monomers
Monomer('C8', ['b'])
Monomer('Bid', ['b', 'S'], {'S':['u', 't']})

# input the parameter values
Parameter('kf', 1.0e-07)
Parameter('kr', 1.0e-03)
Parameter('kc', 1.0)

# OLD RULES
# Rule('C8_Bid_bind', C8(b=None) + Bid(b=None, S=None) | C8(b=1) % Bid(b=1, S=None), *[kf, kr])
# Rule('tBid_from_C8Bid', C8(b=1) % Bid(b=1, S='u') >> C8(b=None) + Bid(b=None, S='t'), kc)
#
# NEW RULES
# Catalysis
catalyze(C8, Bid, 'S', 'u', 't', kf, kr, kc)


# initial conditions
Parameter('C8_0', 1000)
Parameter('Bid_0', 10000)
Initial(C8(b=None), C8_0)
Initial(Bid(b=None, S='u'), Bid_0)

# Observables
Observable('obsC8', C8(b=None))
Observable('obsBid', Bid(b=None, S='u'))
Observable('obstBid', Bid(b=None, S='t'))




#!/usr/bin/env python
"""
Usage
=====

Usage: ``python -m pysb.tools.render_reactions mymodel.py > mymodel.dot``

Renders the reactions produced by a model into the "dot" graph format which can
be visualized with Graphviz.

To create a PDF from the .dot file, use the "dot" command from Graphviz::

    dot mymodel.dot -T pdf -O

This will create mymodel.dot.pdf. You can also change the "dot" command to one
of the other Graphviz drawing tools for a different type of layout. Note that
you can pipe the output of render_reactions straight into Graphviz without
creating an intermediate .dot file, which is especially helpful if you are
making continuous changes to the model and need to visualize your changes
repeatedly::

    python -m pysb.tools.render_reactions mymodel.py | dot -T pdf -o mymodel.pdf

Note that some PDF viewers will auto-reload a changed PDF, so you may not even
need to manually reopen it every time you rerun the tool.

Output for Robertson example model
==================================

The Robertson example model (in ``pysb.examples.robertson``) contains
the following three reactions:

* A -> B
* B + B -> B + C
* C + B -> C + A

The reaction network diagram for this system as generated by this module and
rendered using ``dot`` is shown below:

.. image:: robertson_reactions.png
   :align: center
   :alt: Reaction network for pysb.examples.robertson

Circular nodes (``r0``, ``r1`` and ``r2``) indicate reactions; square nodes
(``A()``, ``B()`` and ``C()``) indicate species. Incoming arrows from a species
node to a reaction node indicate that the species is a reactant; outgoing
arrows from a reaction node to a species node indicate that the species is a
product. A hollow diamond-tipped arrow from a species to a reaction indicates
that the species is involved as both a reactant and a product, i.e., it serves
as a "modifier" (enzyme or catalyst).

"""

from __future__ import print_function
import pysb
import pysb.bng
import re
import sys
import os
try:
    import pygraphviz
except ImportError:
    pygraphviz = None


def run(model):
    """
    Render the reactions produced by a model into the "dot" graph format.

    Parameters
    ----------
    model : pysb.core.Model
        The model to render.

    Returns
    -------
    string
        The dot format output.
    """
    if pygraphviz is None:
        raise ImportError('pygraphviz library is required to run this '
                          'function')

    pysb.bng.generate_equations(model)

    graph = pygraphviz.AGraph(directed=True, rankdir="LR")
    ic_species = [cp for cp, parameter in model.initial_conditions]
    for i, cp in enumerate(model.species):
        species_node = 's%d' % i
        slabel = re.sub(r'% ', r'%\\l', str(cp))
        slabel += '\\l'
        color = "#ccffcc"
        # color species with an initial condition differently
        if len([s for s in ic_species if s.is_equivalent_to(cp)]):
            color = "#aaffff"
        graph.add_node(species_node,
                       label=slabel,
                       shape="Mrecord",
                       fillcolor=color, style="filled", color="transparent",
                       fontsize="12",
                       margin="0.06,0")
    for i, reaction in enumerate(model.reactions_bidirectional):
        reaction_node = 'r%d' % i
        graph.add_node(reaction_node,
                       label=reaction_node,
                       shape="circle",
                       fillcolor="lightgray", style="filled", color="transparent",
                       fontsize="12",
                       width=".3", height=".3", margin="0.06,0")
        reactants = set(reaction['reactants'])
        products = set(reaction['products'])
        modifiers = reactants & products
        reactants = reactants - modifiers
        products = products - modifiers
        attr_reversible = {'dir': 'both', 'arrowtail': 'empty'} if reaction['reversible'] else {}
        for s in reactants:
            r_link(graph, s, i, **attr_reversible)
        for s in products:
            r_link(graph, s, i, _flip=True, **attr_reversible)
        for s in modifiers:
            r_link(graph, s, i, arrowhead="odiamond")
    return graph.string()

def r_link(graph, s, r, **attrs):
    nodes = ('s%d' % s, 'r%d' % r)
    if attrs.get('_flip'):
        del attrs['_flip']
        nodes = reversed(nodes)
    attrs.setdefault('arrowhead', 'normal')
    graph.add_edge(*nodes, **attrs)


usage = __doc__
usage = usage[1:]  # strip leading newline

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        print(usage, end=' ')
        exit()
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module
        # which we use, there will be trouble
        # (use the imp package and import as some safe name?)
        model_module = __import__(model_name)
    except Exception as e:
        print("Error in model script:\n")
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print(run(model))






"""
Module containing a class for exporting a PySB model to SBML.

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.
"""

# FIXME this should use libsbml if available

import pysb
import pysb.bng
from pysb.export import Exporter
import sympy
from sympy.printing.mathml import MathMLPrinter
import itertools
import textwrap
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

def indent(text, n=0):
    """Re-indent a multi-line string, stripping leading newlines and trailing
    spaces."""
    text = text.lstrip('\n')
    text = textwrap.dedent(text)
    lines = text.split('\n')
    lines = [' '*n + l for l in lines]
    text = '\n'.join(lines)
    text = text.rstrip(' ')
    return text

class MathMLContentPrinter(MathMLPrinter):
    """Prints an expression to MathML without presentation markup."""
    def _print_Symbol(self, sym):
        ci = self.dom.createElement(self.mathml_tag(sym))
        ci.appendChild(self.dom.createTextNode(sym.name))
        return ci

def print_mathml(expr, **settings):
    return MathMLContentPrinter(settings).doprint(expr)

def format_single_annotation(annotation):
    return '''<rdf:li rdf:resource="%s" />''' % annotation.object

def get_annotation_preamble(meta_id):
    return indent('''
      <annotation>
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:bqbiol="http://biomodels.net/biology-qualifiers/" xmlns:bqmodel="http://biomodels.net/model-qualifiers/">
              <rdf:Description rdf:about="#%s">
      ''' % meta_id)

def get_annotation_postamble():
    return indent('''
              </rdf:Description>
          </rdf:RDF>
      </annotation>
      ''')

def get_species_annotation(meta_id, cp):
    output = get_annotation_preamble(meta_id)
    model = cp.monomer_patterns[0].monomer.model()
    if len(cp.monomer_patterns) == 1:
        # single monomer
        all_annotations = model.get_annotations(cp.monomer_patterns[0].monomer)
        groups = itertools.groupby(all_annotations, lambda a: a.predicate)
        for predicate, annotations in groups:
            qualifier = 'bqbiol:%s' % predicate
            output += indent('<%s>\n    <rdf:Bag>\n' % qualifier, 12)
            for a in annotations:
                output += indent(format_single_annotation(a) + '\n', 20)
            output += indent('    </rdf:Bag>\n</%s>\n' % qualifier, 12)
    else:
        # complex
        monomers = set(mp.monomer for mp in cp.monomer_patterns)
        annotations = [a for m in monomers for a in model.get_annotations(m)
                       if a.predicate in ('is', 'hasPart')]
        qualifier = 'bqbiol:hasPart'
        output += indent('<%s>\n    <rdf:Bag>\n' % qualifier, 12)
        for a in annotations:
            output += indent(format_single_annotation(a) + '\n', 20)
        output += indent('    </rdf:Bag>\n</%s>\n' % qualifier, 12)
    output += get_annotation_postamble()
    return indent(output, 16)

class SbmlExporter(Exporter):
    """A class for returning the SBML for a given PySB model.

    Inherits from :py:class:`pysb.export.Exporter`, which implements
    basic functionality for all exporters.
    """

    def export(self):
        """Export the SBML for the PySB model associated with the exporter.

        Returns
        -------
        string
            String containing the SBML output.
        """

        output = StringIO()
        pysb.bng.generate_equations(self.model)

        output.write(
            """<?xml version="1.0" encoding="UTF-8"?>
    <sbml xmlns="http://www.sbml.org/sbml/level2" level="2" version="1">
        <model name="%s">""" % self.model.name)

        if self.docstring:
            notes_str = """
            <notes>
                <body xmlns="http://www.w3.org/1999/xhtml">
                    <p>%s</p>
                </body>
            </notes>""" % self.docstring.replace("\n", "<br />\n"+" "*20)
            output.write(notes_str)

        output.write("""
            <listOfCompartments>
                <compartment id="default" name="default" spatialDimensions="3" size="1"/>
            </listOfCompartments>
    """)

        # complexpattern, initial value
        ics = [[s, 0] for s in self.model.species]
        for cp, ic_param in self.model.initial_conditions:
            ics[self.model.get_species_index(cp)][1] = ic_param.value
        output.write("        <listOfSpecies>\n")
        for i, (cp, value) in enumerate(ics):
            id = '__s%d' % i
            metaid = 'metaid_%d' % i
            name = str(cp).replace('% ', '._br_')  # CellDesigner does something weird with % in names
            output.write('            <species id="%s" metaid="%s" name="%s" compartment="default" initialAmount="%.17g">\n'
                         % (id, metaid, name, value));
            output.write(get_species_annotation(metaid, cp))
            output.write('            </species>\n')
        output.write("        </listOfSpecies>\n")

        output.write("        <listOfParameters>\n")
        for i, param in enumerate(self.model.parameters_rules()):
            output.write('            <parameter id="%s" metaid="metaid_%s" name="%s" value="%.17g"/>\n'
                         % (param.name, param.name, param.name, param.value));
        output.write("        </listOfParameters>\n")

        output.write("        <listOfReactions>\n")
        for i, reaction in enumerate(self.model.reactions_bidirectional):
            reversible = str(reaction['reversible']).lower()
            output.write('            <reaction id="r%d" metaid="metaid_r%d" name="r%d" reversible="%s">\n'
                         % (i, i, i, reversible));
            output.write('                <listOfReactants>\n');
            for species in reaction['reactants']:
                output.write('                    <speciesReference species="__s%d"/>\n' % species)
            output.write('                </listOfReactants>\n');
            output.write('                <listOfProducts>\n');
            for species in reaction['products']:
                output.write('                    <speciesReference species="__s%d"/>\n' % species)
            output.write('                </listOfProducts>\n');
            mathml = '<math xmlns="http://www.w3.org/1998/Math/MathML">' \
                + print_mathml(reaction['rate']) + '</math>'
            output.write('                <kineticLaw>' + mathml + '</kineticLaw>\n');
            output.write('            </reaction>\n');
        output.write("        </listOfReactions>\n")

        output.write("    </model>\n</sbml>\n")
        return output.getvalue()





from pysb.examples.michment import model
from pysb.simulator import ScipyOdeSimulator
import numpy as np
import matplotlib.pyplot as plt
import itertools

# Construct initials dict
mult = np.linspace(0.8, 1.2, 11)
matrix = [mult*ic[1].value for ic in model.initial_conditions]
initials = {}
for i,ic in enumerate(model.initial_conditions):
    initials[ic[0]] = []
    for tup in itertools.product(*matrix): # Cartesian product
        initials[ic[0]].append(tup[i])

tspan = np.linspace(0, 50, 501)
sim = ScipyOdeSimulator(model, tspan, verbose=False)
trajectories = sim.run(initials=initials).all

x = np.array([tr['Product'] for tr in trajectories]).T

plt.plot(tspan, x.mean(axis=1), 'b', lw=3, label="Product")
plt.plot(tspan, x.max(axis=1), 'b--', lw=2, label="min/max")
plt.plot(tspan, x.min(axis=1), 'b--', lw=2)
plt.xlabel('time')
plt.ylabel('concentration')
plt.legend(loc='upper left')

plt.show()



# import the pysb module and all its methods and functions
from pysb import *

# instantiate a model
Model()

# declare monomers
Monomer('C8', ['b'])
Monomer('Bid', ['b', 'S'], {'S':['u', 't']})

# input the parameter values
Parameter('kf', 1.0e-07)
Parameter('kr', 1.0e-03)
Parameter('kc', 1.0)

# now input the rules
Rule('C8_Bid_bind', C8(b=None) + Bid(b=None, S=None) | C8(b=1) % Bid(b=1, S=None), *[kf, kr])
Rule('tBid_from_C8Bid', C8(b=1) % Bid(b=1, S='u') >> C8(b=None) + Bid(b=None, S='t'), kc)

# initial conditions
Parameter('C8_0', 1000)
Parameter('Bid_0', 10000)
Initial(C8(b=None), C8_0)
Initial(Bid(b=None, S='u'), Bid_0)

# Observables
Observable('obsC8', C8(b=None))
Observable('obsBid', Bid(b=None, S='u'))
Observable('obstBid', Bid(b=None, S='t'))




import warnings
import numpy as np
from nose.plugins.attrib import attr
from pysb.examples.tyson_oscillator import model
from pysb.simulator.cupsoda import CupSodaSimulator, run_cupsoda
from nose.tools import raises
import os


@attr('gpu')
class TestCupSODASimulatorSingle(object):
    def setUp(self):
        self.n_sims = 50
        self.tspan = np.linspace(0, 500, 101)
        self.solver = CupSodaSimulator(model, tspan=self.tspan, verbose=False,
                                       integrator_options={'atol': 1e-12,
                                                           'rtol': 1e-12,
                                                           'max_steps': 20000})
        len_model_species = len(model.species)
        y0 = np.zeros((self.n_sims, len_model_species))
        for ic in model.initial_conditions:
            for j in range(len_model_species):
                if str(ic[0]) == str(model.species[j]):
                    y0[:, j] = ic[1].value
                    break
        self.y0 = y0

    def test_use_of_volume(self):
        # Initial concentrations
        self.solver.run(initials=self.y0)
        print(self.solver.vol)
        assert self.solver.vol is None
        self.solver.vol = 1e-20
        assert self.solver.vol == 1e-20

    def test_integrator_options(self):
        assert self.solver.opts['atol'] == 1e-12
        assert self.solver.opts['rtol'] == 1e-12
        assert self.solver.opts['max_steps'] == 20000

    def test_arguments(self):
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Neither 'param_values' nor "
                                              "'initials' were supplied.")
            self.solver.run(param_values=None, initials=None)

    def test_memory_usage(self):
        assert self.solver.opts['memory_usage'] == 'sharedconstant'
        self.solver.run(initials=self.y0)  # memory_usage='sharedconstant'
        self.solver.opts['memory_usage'] = 'global'
        self.solver.run(initials=self.y0)
        self.solver.opts['memory_usage'] = 'shared'
        self.solver.run(initials=self.y0)

    def test_n_blocks(self):
        print(self.solver.n_blocks)
        self.solver.n_blocks = 128
        assert self.solver.n_blocks == 128
        self.solver.run(initials=self.y0)

    @raises(ValueError)
    def test_set_nblocks_str(self):
        self.solver.n_blocks = 'fail'

    @raises(ValueError)
    def test_set_nblocks_0(self):
        self.solver.n_blocks = 0

    def test_run_tyson(self):
        # Rate constants
        len_parameters = len(model.parameters)
        param_values = np.ones((self.n_sims, len_parameters))
        for j in range(len_parameters):
            param_values[:, j] *= model.parameters[j].value
        simres = self.solver.run(initials=self.y0)
        print(simres.observables)
        self.solver.run(param_values=None, initials=self.y0)
        self.solver.run(param_values=param_values, initials=self.y0)
        self.solver.run(param_values=param_values, initials=self.y0)

    def test_verbose(self):
        solver = CupSodaSimulator(model, tspan=self.tspan, verbose=True,
                                  vol=1e-5,
                                  integrator_options={'atol': 1e-12,
                                                      'rtol': 1e-12,
                                                      'max_steps': 20000})
        solver.run()

    def test_run_cupsoda_instance(self):
        run_cupsoda(model, tspan=self.tspan)



from __future__ import print_function as _
from pysb import ComponentSet
import pysb.core
import inspect
import numpy
import io
import networkx as nx
try:
    basestring
except NameError:
    basestring = str

__all__ = ['alias_model_components', 'rules_using_parameter', 'read_dot']


def alias_model_components(model=None):
    """Make all model components visible as symbols in the caller's global namespace"""
    if model is None:
        model = pysb.core.SelfExporter.default_model
    caller_globals = inspect.currentframe().f_back.f_globals
    components = dict((c.name, c) for c in model.all_components())
    caller_globals.update(components)


def rules_using_parameter(model, parameter):
    """Return a ComponentSet of rules in the model which make use of the given parameter"""
    if not isinstance(parameter, pysb.core.Parameter):
        # Try getting the parameter by name
        parameter = model.parameters.get(parameter)
    cset = ComponentSet()
    for rule in model.rules:
        if rule.rate_forward is parameter or rule.rate_reverse is parameter:
            cset.add(rule)
    return cset


def synthetic_data(model, tspan, obs_list=None, sigma=0.1):
    #from pysb.integrate import odesolve
    from pysb.integrate import Solver
    solver = Solver(model, tspan)
    solver.run()

    # Sample from a normal distribution with variance sigma and mean 1
    # (randn generates a matrix of random numbers sampled from a normal
    # distribution with mean 0 and variance 1)
    # 
    # Note: This modifies yobs_view (the view on yobs) so that the changes 
    # are reflected in yobs (which is returned by the function). Since a new
    # Solver object is constructed for each function invocation this does not
    # cause problems in this case.
    solver.yobs_view *= ((numpy.random.randn(*solver.yobs_view.shape) * sigma) + 1)
    return solver.yobs


def get_param_num(model, name):
    for i in range(len(model.parameters)):
        if model.parameters[i].name == name:
            print(i, model.parameters[i])
            break
    return i


def write_params(model,paramarr, name=None):
    """ write the parameters and values to a csv file
    model: a model object
    name: a string with the name for the file, or None to return the content
    """
    if name is not None:
        fobj = open(name, 'w')
    else:
        fobj = io.StringIO()
    for i in range(len(model.parameters)):
        fobj.write("%s, %.17g\n"%(model.parameters[i].name, paramarr[i]))
    if name is None:
        return fobj.getvalue()


def update_param_vals(model, newvals):
    """update the values of model parameters with the values from a dict. 
    the keys in the dict must match the parameter names
    """
    update = []
    noupdate = []
    for i in model.parameters:
        if i.name in newvals:
            i.value = newvals[i.name]
            update.append(i.name)
        else:
            noupdate.append(i.name)
    return update, noupdate


def load_params(fname):
    """load the parameter values from a csv file, return them as dict.
    """
    parmsff = {}
    # FIXME: This might fail if a parameter name is larger than 50 characters.
    # FIXME: Maybe do this with the csv module instead?
    temparr = numpy.loadtxt(fname, dtype=([('a','S50'),('b','f8')]), delimiter=',') 
    for i in temparr:
        parmsff[i[0]] = i[1]
    return parmsff


def read_dot(filename):
    """ Read a graphviz dot file using pydot

    Parameters
    ----------
    filename: str
        A DOT (graphviz) filename

    Returns
    -------
    MultiGraph
        A networkx MultiGraph file
    """
    try:
        import pydot
        pydot_graph = pydot.graph_from_dot_file(filename)[0]
        return _from_pydot(pydot_graph)
    except ImportError:
        raise ImportError('Importing graphviz files requires the pydot '
                          'library')


def _from_pydot(P):
    """Return a NetworkX graph from a Pydot graph.

    Using this patched version until networkx issue is resolved, which fixes
    .dot file support using PyDot:
    https://github.com/networkx/networkx/issues/2832

    Parameters
    ----------

    P : Pydot graph
      A graph created with Pydot

    Returns
    -------
    G : NetworkX multigraph
        A MultiGraph or MultiDiGraph.

    Examples
    --------
    >>> K5 = nx.complete_graph(5)
    >>> A = nx.nx_pydot.to_pydot(K5)
    >>> G = nx.nx_pydot.from_pydot(A) # return MultiGraph

    # make a Graph instead of MultiGraph
    >>> G = nx.Graph(nx.nx_pydot.from_pydot(A))

    """
    if P.get_strict(None):  # pydot bug: get_strict() shouldn't take argument
        multiedges = False
    else:
        multiedges = True

    if P.get_type() == 'graph':  # undirected
        if multiedges:
            N = nx.MultiGraph()
        else:
            N = nx.Graph()
    else:
        if multiedges:
            N = nx.MultiDiGraph()
        else:
            N = nx.DiGraph()

    # assign defaults
    name = P.get_name().strip('"')
    if name != '':
        N.name = name

    # add nodes, attributes to N.node_attr
    for p in P.get_node_list():
        n = p.get_name().strip('"')
        if n in ('node', 'graph', 'edge'):
            continue
        N.add_node(n, **{k: v.strip('"')
                         for k, v in p.get_attributes().items()})

    # add edges
    for e in P.get_edge_list():
        u = e.get_source()
        v = e.get_destination()
        attr = {k: v.strip('"') for k, v in e.get_attributes().items()}
        s = []
        d = []

        if isinstance(u, basestring):
            s.append(u.strip('"'))
        else:
            for unodes in u['nodes']:
                s.append(unodes.strip('"'))

        if isinstance(v, basestring):
            d.append(v.strip('"'))
        else:
            for vnodes in v['nodes']:
                d.append(vnodes.strip('"'))

        for source_node in s:
            for destination_node in d:
                N.add_edge(source_node, destination_node, **attr)

    # add default attributes for graph, nodes, edges
    pattr = {k: v.strip('"') for k, v in P.get_attributes().items()}
    if pattr:
        N.graph['graph'] = pattr
    try:
        N.graph['node'] = P.get_node_defaults()[0]
    except:  # IndexError,TypeError:
        pass  # N.graph['node']={}
    try:
        N.graph['edge'] = P.get_edge_defaults()[0]
    except:  # IndexError,TypeError:
        pass  # N.graph['edge']={}
    return N



import os
from itertools import product
import collections
import matplotlib.pyplot as plt
from matplotlib import gridspec
import numpy as np
from pysb.bng import generate_equations
import pysb.simulator.base
from pysb.logging import get_logger


class InitialsSensitivity(object):
    """
    Pairwise sensitivity analysis of initial conditions

    This class calculates the sensitivity of a specified model Observable to
    changes in pairs of initial species concentrations. The results are
    stored in matrices described in Attributes.

    .. warning::
        The interface for this class is considered experimental and may
        change without warning as PySB is updated.

    Parameters
    ----------
    solver : pysb.simulator.Simulator
        Simulator instance used to perform models. Must be initialized with
        tspan argument set.
    values_to_sample : vector_like
        Values to sample for each initial concentration of the
        model.parameters values.
    objective_function : function
        A function that returns a scalar value. Used to calculate fraction
        of changed that is used for calculating sensitivity. See Example.
    observable : str
        Observable used in the objective_function.

    Attributes
    ----------
    b_matrix: numpy.ndarray
        Matrix of 2-tuples containing (perturbation, species index)
    b_prime_matrix: numpy.ndarray
        Same as b_matrix, only where one of the species concentrations is
        unchanged (i.e. with the single variable contribution removed)
    p_matrix: numpy.ndarray
        Pairwise sensitivity matrix
    p_prime_matrix: numpy.ndarray
        Normalized pairwise sensitivity matrix (in the sense that it
        contains changes from the baseline, unperturbed case)

    References
    ----------
    Harris et al. Bioinformatics (2017), under review.

    Examples
    --------
    Sensitivity analysis on the Tyson cell cycle model

    >>> from pysb.examples.tyson_oscillator import model
    >>> import numpy as np
    >>> from pysb.simulator.scipyode import ScipyOdeSimulator
    >>> np.set_printoptions(precision=4)
    >>> tspan=np.linspace(0, 200, 201)
    >>> observable = 'Y3'
    >>> values_to_sample = [.8, 1.2]
    >>> def obj_func_cell_cycle(out):
    ...     timestep = tspan[:-1]
    ...     y = out[:-1] - out[1:]
    ...     freq = 0
    ...     local_times = []
    ...     prev = y[0]
    ...     for n in range(1, len(y)):
    ...         if y[n] > 0 > prev:
    ...             local_times.append(timestep[n])
    ...             freq += 1
    ...         prev = y[n]
    ...     local_times = np.array(local_times)
    ...     local_freq = np.average(local_times)/len(local_times)*2
    ...     return local_freq
    >>> solver = ScipyOdeSimulator(model, tspan)
    >>> sens = InitialsSensitivity(\
            values_to_sample=values_to_sample,\
            observable=observable,\
            objective_function=obj_func_cell_cycle,\
            solver=solver\
        )
    >>> print(sens.b_matrix)
    [[((0.8, 'cdc0'), (0.8, 'cdc0')) ((0.8, 'cdc0'), (1.2, 'cdc0'))
      ((0.8, 'cdc0'), (0.8, 'cyc0')) ((0.8, 'cdc0'), (1.2, 'cyc0'))]
     [((1.2, 'cdc0'), (0.8, 'cdc0')) ((1.2, 'cdc0'), (1.2, 'cdc0'))
      ((1.2, 'cdc0'), (0.8, 'cyc0')) ((1.2, 'cdc0'), (1.2, 'cyc0'))]
     [((0.8, 'cyc0'), (0.8, 'cdc0')) ((0.8, 'cyc0'), (1.2, 'cdc0'))
      ((0.8, 'cyc0'), (0.8, 'cyc0')) ((0.8, 'cyc0'), (1.2, 'cyc0'))]
     [((1.2, 'cyc0'), (0.8, 'cdc0')) ((1.2, 'cyc0'), (1.2, 'cdc0'))
      ((1.2, 'cyc0'), (0.8, 'cyc0')) ((1.2, 'cyc0'), (1.2, 'cyc0'))]]
    >>> sens.run()
    >>> print(sens.p_matrix)#doctest: +NORMALIZE_WHITESPACE
    [[ 0.      0.      5.0243 -4.5381]
     [ 0.      0.      5.0243 -4.5381]
     [ 5.0243  5.0243  0.      0.    ]
     [-4.5381 -4.5381  0.      0.    ]]

    >>> print(sens.p_prime_matrix) #doctest: +NORMALIZE_WHITESPACE
     [[ 0.      0.      5.0243 -4.5381]
      [ 0.      0.      5.0243 -4.5381]
      [ 0.      0.      0.      0.    ]
      [ 0.      0.      0.      0.    ]]
    >>> print(sens.p_matrix - sens.p_prime_matrix) \
            #doctest: +NORMALIZE_WHITESPACE
     [[ 0.      0.      0.      0.    ]
      [ 0.      0.      0.      0.    ]
      [ 5.0243  5.0243  0.      0.    ]
      [-4.5381 -4.5381  0.      0.    ]]
    >>> sens.create_boxplot_and_heatplot() #doctest: +SKIP
    """

    def __init__(self, solver, values_to_sample, objective_function,
                 observable):
        self._model = solver.model
        self._logger = get_logger(__name__, model=self._model)
        self._logger.info('%s created for observable %s' % (
            self.__class__.__name__, observable))
        generate_equations(self._model)
        self._ic_params_of_interest_cache = None
        self._values_to_sample = values_to_sample
        if solver is None or not isinstance(solver,
                                            pysb.simulator.base.Simulator):
            raise(TypeError, "solver must be a pysb.simulator object")
        self._solver = solver
        self.objective_function = objective_function
        self.observable = observable

        # Outputs
        self.b_matrix = []
        self.b_prime_matrix = []
        self.p_prime_matrix = np.zeros(self._size_of_matrix)
        self.p_matrix = np.zeros(self._size_of_matrix)

        self._original_initial_conditions = np.zeros(len(self._model.species))
        self._index_of_species_of_interest = self._create_index_of_species()
        self.simulation_initials = self._setup_simulations()
        # Stores the objective function value for the original unperturbed
        # model
        self._objective_fn_standard = None

    @property
    def _ic_params_of_interest(self):
        if self._ic_params_of_interest_cache is None:
            self._ic_params_of_interest_cache = list(
                (i[1].name for i in self._model.initial_conditions))
            self._ic_params_of_interest_cache = \
                sorted(self._ic_params_of_interest_cache)

        return self._ic_params_of_interest_cache

    @property
    def _n_species(self):
        return len(self._ic_params_of_interest)

    @property
    def _n_sam(self):
        return len(self._values_to_sample)

    @property
    def _nm(self):
        return self._n_species * self._n_sam

    @property
    def _size_of_matrix(self):
        return self._nm ** 2

    @property
    def _shape_of_matrix(self):
        return self._nm, self._nm

    @property
    def sensitivity_multiset(self):
        """
        Sensitivity analysis multiset (also called "Q" matrix)

        Returns
        -------
        list
            List of lists containing the sensitivity analysis multiset
        """
        sens_ij_nm = []
        sens_matrix = self.p_matrix - self.p_prime_matrix

        # separate each species sensitivity
        for j in range(0, self._nm, self._n_sam):
            per_species1 = []
            for i in range(0, self._nm, self._n_sam):
                if i != j:
                    tmp = sens_matrix[j:j + self._n_sam,
                          i:i + self._n_sam].copy()
                    per_species1.append(tmp)
            sens_ij_nm.append(per_species1)
        return sens_ij_nm

    def _calculate_objective(self, function_value):
        """
        Calculate fraction of change for obj value and standard

        Parameters
        ----------
        function_value : scalar
            scalar value provided by objective function
        """
        return (self.objective_function(function_value)
                - self._objective_fn_standard) / \
               self._objective_fn_standard * 100.

    def run(self, save_name=None, out_dir=None):
        """
        Run sensitivity analysis

        Parameters
        ----------
        save_name : str, optional
            prefix of saved files
        out_dir : str, optional
            location to save output if required
        """
        if out_dir is not None:
            if not os.path.exists(out_dir):
                os.mkdir(out_dir)
        self._solver.initials = None
        self._solver.param_values = None
        sim_results = self._solver.run(param_values=None, initials=None)
        self._objective_fn_standard = self.objective_function(
            np.array(sim_results.observables[self.observable]))

        traj = self._solver.run(initials=self.simulation_initials)
        sensitivity_output = np.array(traj.observables)[self.observable].T
        p_matrix = np.zeros(self._size_of_matrix)
        p_prime_matrix = np.zeros(self._size_of_matrix)
        counter = 0
        # places values in p matrix that are unique
        for i in range(len(p_matrix)):
            if i in self.b_index:
                tmp = self._calculate_objective(sensitivity_output[:, counter])
                p_matrix[i] = tmp
                counter += 1
        # places values in p matrix that are duplicated
        for i in range(len(p_matrix)):
            if i in self.b_prime_not_in_b:
                tmp = self._calculate_objective(
                    sensitivity_output[:, self.b_prime_not_in_b[i]])
                p_prime_matrix[i] = tmp
            elif i in self.b_prime_in_b:
                p_prime_matrix[i] = p_matrix[self.b_prime_in_b[i]]
        # Reshape
        p_matrix = p_matrix.reshape(self._shape_of_matrix)
        # Project the mirrored image
        self.p_matrix = p_matrix + p_matrix.T
        self.p_prime_matrix = p_prime_matrix.reshape(self._shape_of_matrix)

        # save output if desired
        if save_name is not None:
            if out_dir is None:
                out_dir = '.'
            p_name = os.path.join(out_dir, '{}_p_matrix.csv'.format(save_name))
            p_prime_name = os.path.join(
                out_dir, '{}_p_prime_matrix.csv'.format(save_name))
            self._logger.debug("Saving p matrix and p' matrix to {} and {}".
                               format(p_name, p_prime_name))

            np.savetxt(p_name, self.p_matrix)
            np.savetxt(p_prime_name, self.p_prime_matrix)

    def _create_index_of_species(self):
        """ create dictionary of initial conditions by index """
        index_of_init_condition = {}
        for ic_sp, ic_param in self._model.initial_conditions:
            for sp_idx, sp in enumerate(self._model.species):
                if ic_sp.is_equivalent_to(sp):
                    self._original_initial_conditions[sp_idx] = ic_param.value
                    if ic_param.name in self._ic_params_of_interest:
                        index_of_init_condition[ic_param.name] = sp_idx
        index_of_species_of_interest = collections.OrderedDict(
            sorted(index_of_init_condition.items()))
        return index_of_species_of_interest

    def _g_function(self):
        """ creates sample matrix, index of samples values, and shows bij """
        counter = -1
        sampled_values_index = set()
        bij_unique = dict()
        sampling_matrix = np.zeros(
            (self._size_of_matrix, len(self._model.species)))
        sampling_matrix[:, :] = self._original_initial_conditions
        matrix = self.b_matrix
        for j in range(len(matrix)):
            for i in matrix[j, :]:
                sigma_i, index_i = i[0]
                sigma_j, index_j = i[1]
                s_1 = str(i[0]) + str(i[1])
                s_2 = str(i[1]) + str(i[0])
                counter += 1

                if index_i == index_j:
                    continue
                elif s_1 in bij_unique or s_2 in bij_unique:
                    continue
                else:
                    x = self._index_of_species_of_interest[index_i]
                    y = self._index_of_species_of_interest[index_j]
                    sampling_matrix[counter, x] *= sigma_i
                    sampling_matrix[counter, y] *= sigma_j
                    bij_unique[s_1] = counter
                    bij_unique[s_2] = counter
                    sampled_values_index.add(counter)

        return sampling_matrix, sampled_values_index, bij_unique

    def _setup_simulations(self):
        """
        Create initial conditions matrix for sensitivity analysis

        Returns
        -------
        numpy.ndarray
            Matrix of initial conditions
        """
        # create matrix (cartesian product of sample vals vs index of species
        a_matrix = cartesian_product(self._values_to_sample,
                                     self._index_of_species_of_interest)
        # reshape to flatten
        a_matrix = a_matrix.T.reshape(self._n_sam * self._n_species)
        # creates matrix b
        self.b_matrix = cartesian_product(a_matrix, a_matrix)

        # create matrix a'
        a_prime = cartesian_product(np.ones(self._n_sam),
                                    self._index_of_species_of_interest)
        a_prime = a_prime.T.reshape(self._n_sam * self._n_species)

        # creates matrix b prime
        self.b_prime_matrix = cartesian_product(a_prime, a_matrix)

        b_to_run, self.b_index, in_b = self._g_function()

        n_b_index = len(self.b_index)

        b_prime = np.zeros((self._size_of_matrix, len(self._model.species)))
        b_prime[:, :] = self._original_initial_conditions
        counter = -1

        bp_not_in_b_raw = set()
        bp_dict = dict()
        bp_not_in_b_dict = dict()
        bp_not_in_b_visited = dict()
        new_sim_counter = -1
        # checks for and removes duplicates of simulations initial conditions
        for j in range(len(self.b_prime_matrix)):
            for i in self.b_prime_matrix[j, :]:
                sigma_i, index_i = i[0]
                sigma_j, index_j = i[1]
                s_1 = str(i[0]) + str(i[1])
                counter += 1
                # no need for doing if same index
                if index_i == index_j:
                    continue
                # pointing to same simulation if already in b
                elif s_1 in in_b:
                    bp_dict[counter] = in_b[s_1]
                elif s_1 in bp_not_in_b_visited:
                    bp_not_in_b_dict[counter] = bp_not_in_b_visited[s_1]
                else:
                    new_sim_counter += 1
                    x = self._index_of_species_of_interest[index_i]
                    y = self._index_of_species_of_interest[index_j]
                    b_prime[new_sim_counter, x] *= sigma_i
                    b_prime[new_sim_counter, y] *= sigma_j
                    bp_not_in_b_visited[s_1] = new_sim_counter + n_b_index
                    bp_not_in_b_dict[counter] = new_sim_counter + n_b_index
                    bp_not_in_b_raw.add(new_sim_counter)

        self.b_prime_in_b = bp_dict
        self.b_prime_not_in_b = bp_not_in_b_dict
        x = b_to_run[list(self.b_index)]
        y = b_prime[list(bp_not_in_b_raw)]
        simulations = np.vstack((x, y))
        self._logger.debug("Number of simulations to run = %s" % len(
                           simulations))
        return simulations

    def create_plot_p_h_pprime(self, save_name=None, out_dir=None, show=False):
        """
        Plot of P, H(B), and P'

        See :class:`InitialsSensitivity` for descriptions of these matrices

        Parameters
        ----------
        save_name : str, optional
            name to save figure as
        out_dir : str, optional
            location to save figure
        show : bool
            show the plot if True

        Returns
        -------
        matplotlib.figure.Figure
            The matplotlib figure object for further adjustments, if required
        """
        colors = 'seismic'
        sens_matrix = self.p_matrix - self.p_prime_matrix
        v_max = max(np.abs(self.p_matrix.min()), self.p_matrix.max())
        v_min = -1 * v_max

        fig = plt.figure(figsize=(12, 6))
        ax1 = fig.add_subplot(131)
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        ax1.imshow(self.p_matrix, interpolation='nearest', origin='upper',
                   cmap=plt.get_cmap(colors), vmin=v_min, vmax=v_max,
                   extent=[0, self._nm, 0, self._nm])

        ax2.imshow(self.p_prime_matrix, interpolation='nearest',
                   origin='upper', cmap=plt.get_cmap(colors),
                   vmin=v_min, vmax=v_max, extent=[0, self._nm, 0, self._nm])

        ax3.imshow(sens_matrix, interpolation='nearest', origin='upper',
                   cmap=plt.get_cmap(colors), vmin=v_min, vmax=v_max)

        for ax in (ax1, ax2, ax3):
            ax.set_xticks([])
            ax.set_yticks([])

        ax1.set_title('P', fontsize=22)
        ax2.set_title('H(B\')', fontsize=22)
        ax3.set_title('P\' = P - H(B\')', fontsize=22)
        fig.subplots_adjust(wspace=0, hspace=0.0)
        if save_name is not None:
            if out_dir is None:
                out_dir = '.'
            if not os.path.exists(out_dir):
                os.mkdir(out_dir)
            fig.savefig(os.path.join(out_dir,
                                     '{}_P_H_P_prime.png'.format(save_name)),
                        bbox_inches='tight')
        if show:
            plt.show()
        plt.close()

        return fig

    def create_individual_pairwise_plots(self, save_name=None, out_dir=None,
                                         show=False):
        """
        Single plot containing heat plot of each specie pair

        Parameters
        ----------
        save_name : str, optional
            name ot save figure as
        out_dir : str, optional
            output directory
        show : bool
            show figure

        Returns
        -------
        matplotlib.figure.Figure
            The matplotlib figure object for further adjustments, if required
        """
        colors = 'seismic'
        sens_matrix = self.p_matrix - self.p_prime_matrix
        v_max = max(np.abs(self.p_matrix.min()), self.p_matrix.max())
        v_min = -1 * v_max
        fig = plt.figure(figsize=(self._n_species + 6, self._n_species + 6))
        gs = gridspec.GridSpec(self._n_species, self._n_species)
        # creates a plot of each species vs each species
        # adds space between plots so you can zoom in on output pairs
        for n, j in enumerate(range(0, self._nm, self._n_sam)):
            for m, i in enumerate(range(0, self._nm, self._n_sam)):
                ax2 = plt.subplot(gs[n, m])
                if n == 0:
                    ax2.set_xlabel(self._ic_params_of_interest[m], fontsize=20)
                    ax2.xaxis.set_label_position('top')
                if m == 0:
                    ax2.set_ylabel(self._ic_params_of_interest[n], fontsize=20)
                plt.xticks([])
                plt.yticks([])
                if i != j:
                    tmp = sens_matrix[j:j + self._n_sam,
                          i:i + self._n_sam].copy()
                    ax2.imshow(tmp, interpolation='nearest', origin='upper',
                               cmap=plt.get_cmap(colors), vmin=v_min,
                               vmax=v_max)
                else:
                    ax2.imshow(np.zeros((self._n_sam, self._n_sam)),
                               interpolation='nearest', origin='upper',
                               cmap=plt.get_cmap(colors), vmin=v_min,
                               vmax=v_max)
        plt.tight_layout()

        if save_name is not None:
            if out_dir is None:
                out_dir = '.'
            if not os.path.exists(out_dir):
                os.mkdir(out_dir)
            plt.savefig(os.path.join(out_dir,
                                     '{}_subplots.png'.format(save_name)),
                        bbox_inches='tight')
        if show:
            plt.show()
        plt.close()

        return fig

    def create_boxplot_and_heatplot(self, x_axis_label=None, save_name=None,
                                    out_dir=None, show=False):
        """
        Heat map and box plot of sensitivities

        Parameters
        ----------
        x_axis_label : str, optional
            label for x asis
        save_name : str, optional
            name of figure to save
        out_dir : str, option
            output directory to save figures
        show : bool
            Show plot if True

        Returns
        -------
        matplotlib.figure.Figure
            The matplotlib figure object for further adjustments, if required
        """

        colors = 'seismic'
        sens_ij_nm = self.sensitivity_multiset

        # Create heatmap and boxplot of data
        fig = plt.figure(figsize=(14, 10))
        plt.subplots_adjust(hspace=0.1)

        # use gridspec to scale colorbar nicely
        outer = gridspec.GridSpec(2, 1, width_ratios=[1.],
                                  height_ratios=[0.03, 1])

        gs1 = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=outer[0])
        gs2 = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=outer[1],
                                               hspace=.35)
        ax0 = plt.subplot(gs1[0])
        ax1 = plt.subplot(gs2[0])

        # scale the colors to minimum or maximum of p matrix
        v_max = max(np.abs(self.p_matrix.min()), self.p_matrix.max())
        v_min = -1 * v_max

        # create heatmap of sensitivities
        im = ax1.imshow(self.p_matrix, interpolation='nearest', origin='upper',
                        cmap=plt.get_cmap(colors), vmin=v_min,
                        vmax=v_max, extent=[0, self._nm, 0, self._nm])

        shape_label = np.arange(self._n_sam / 2, self._nm, self._n_sam)
        plt.xticks(shape_label, self._ic_params_of_interest,
                   rotation='vertical', fontsize=12)
        plt.yticks(shape_label, reversed(self._ic_params_of_interest),
                   fontsize=12)
        x_ticks = ([i for i in range(0, self._nm, self._n_sam)])
        ax1.set_xticks(x_ticks, minor=True)
        ax1.set_yticks(x_ticks, minor=True)
        plt.grid(True, which='minor', linestyle='--')
        color_bar = plt.colorbar(im, cax=ax0, orientation='horizontal',
                                 use_gridspec=True)
        color_bar.set_label('% change', y=1, labelpad=5)
        color_bar.ax.xaxis.set_label_position('top')
        ticks = np.linspace(v_min, v_max, 5, dtype=int)
        color_bar.set_ticks(ticks)
        color_bar.ax.set_xticklabels(ticks)

        # create boxplot of single parameter sensitivities
        ax2 = plt.subplot(gs2[1])
        x = [np.array(mat).flatten() for mat in sens_ij_nm[::-1]]
        ax2.boxplot(x, vert=False, labels=None, showfliers=True, whis='range')
        ax2.set_xlim(v_min - 2, v_max + 2)
        if x_axis_label is not None:
            ax2.set_xlabel(x_axis_label, fontsize=12)
        plt.setp(ax2, yticklabels=reversed(self._ic_params_of_interest))
        ax2.yaxis.tick_left()
        ax2.set_aspect(1. / ax2.get_data_ratio(), adjustable='box')
        if save_name is not None:
            if out_dir is None:
                out_dir = '.'
            if not os.path.exists(out_dir):
                os.mkdir(out_dir)
            plt.savefig(os.path.join(out_dir, save_name + '.png'),
                        bbox_inches='tight')
            plt.savefig(os.path.join(out_dir, save_name + '.eps'),
                        bbox_inches='tight')
            plt.savefig(os.path.join(out_dir, save_name + '.svg'),
                        bbox_inches='tight')
        if show:
            plt.show()
        plt.close()

        return fig


def cartesian_product(array_1, array_2):
    """
    Cartesian product between two lists

    Parameters
    ----------
    array_1 : list_like
    array_2 : list_like

    Returns
    -------
    out : np.array
        array of shape (len(array_1) x len(array_2))
    """
    a = list(product(array_1, array_2))
    a = np.asarray(a, dtype=','.join(['object'] * len(a[0])))
    return a.reshape(len(array_1), len(array_2))



from pysb import *

Model()
Parameter('atol', 1e5)
Parameter('Btot', 2e5)
Parameter('k1', 3e-7)
Parameter('k2', 1e-4)
Parameter('k2_n', 1e-4/3)
Parameter('k3', 1e-3)
Parameter('k4', 3.5)
Parameter('atol_k1', 1e5 * 3e-7)
Parameter('btol_k3', 2e5 * 1e-3)
Monomer('X')
Monomer('A')
Monomer('B')
Monomer('I')
Initial(X(), Parameter('X_0', 100))
Initial(A(), Parameter('A_0', 1e5))
Initial(B(), Parameter('B_0', 1e5))
Initial(I(), Parameter('I_0', 1))

# Rule('rule1', A() + X() + X() <> A() + X() + X() + X(), k1, k2)
# Rule('rule2', B() <> X() + B(), k3, k4)

Rule('rule3', X() + X() | X() + X() + X(), atol_k1, k2_n)
Rule('rule4', I() | X() + I() , btol_k3, k4)
Observable('X_total', X())
Observable('A_total', A())
Observable('B_total', B())



from pysb.testing import *
from pysb import *
from pysb.kappa import *
import networkx as nx
import sympy
from pysb.core import as_complex_pattern

_KAPPA_SEED = 123456


@with_model
def test_kappa_simulation_results():
    Monomer('A', ['b'])
    Monomer('B', ['b'])
    Initial(A(b=None), Parameter('A_0', 100))
    Initial(B(b=None), Parameter('B_0', 100))
    Rule('A_binds_B', A(b=None) + B(b=None) >> A(b=1) % B(b=1),
         Parameter('kf', 1))
    Rule('A_binds_B_rev', A(b=1) % B(b=1) >> A(b=None) + B(b=None),
         Parameter('kr', 1))
    Observable('AB', A(b=1) % B(b=1))
    npts = 200
    kres = run_simulation(model, time=100, points=npts, seed=_KAPPA_SEED)
    ok_(len(kres['time']) == npts + 1)
    ok_(len(kres['AB']) == npts + 1)
    ok_(kres['time'][0] == 0)
    ok_(sorted(kres['time'])[-1] == 100)


@with_model
def test_kappa_expressions():
    Monomer('A',['site'],{'site': ['u']})
    Parameter('two',2)
    Parameter('kr',0.1)
    Parameter('num_A',1000)
    Expression('kf',1e-5/two)
    Expression('test_sqrt', -1 + sympy.sqrt(1 + two))
    Expression('test_pi', sympy.pi)
    Expression('test_e', sympy.E)
    Expression('test_log', sympy.log(two))
    Expression('test_exp', sympy.exp(two))
    Expression('test_sin', sympy.sin(two))
    Expression('test_cos', sympy.cos(two))
    Expression('test_tan', sympy.tan(two))
    Expression('test_max', sympy.Max(two, kr, 2.0))
    Expression('test_min', sympy.Min(two, kr, 2.0))
    Expression('test_mod', sympy.Mod(10, two))
    Expression('test_piecewise', sympy.Piecewise((0.0, two < 400.0),
                                                 (1.0, True)))
    Initial(A(site=('u')),num_A)
    Rule('dimerize_fwd',
         A(site='u') + A(site='u') >> A(site=('u', 1)) % A(site=('u',1)), kf)
    Rule('dimerize_rev',
         A(site=('u', 1)) % A(site=('u',1)) >>
         A(site='u') + A(site='u'), kr)
    # We need an arbitrary observable here to get a Kappa output file
    Observable('A_obs', A())
    # Accommodates Expression in kappa simulation
    run_simulation(model, time=0)

    Rule('degrade_dimer', A(site=('u', ANY)) >> None, kr)
    Observable('dimer', A(site=('u', ANY)))
    # Accommodates site with explicit state and arbitrary bond
    run_simulation(model, time=0, seed=_KAPPA_SEED)


@with_model
def test_flux_map():
    """Test kappa simulation with flux map (returns tuple with graph)"""
    Monomer('A', ['b'])
    Monomer('B', ['a', 'c'])
    Monomer('C', ['b'])
    Parameter('k', 0.001)
    Rule('A_binds_B', A(b=None) + B(a=None) >> A(b=1) % B(a=1), k)
    Rule('C_binds_B', C(b=None) + B(c=None) >> C(b=1) % B(c=1), k)
    Observable('ABC', A(b=1) % B(a=1, c=2) % C(b=2))
    Initial(A(b=None), Parameter('A_0', 100))
    Initial(B(a=None, c=None), Parameter('B_0', 100))
    Initial(C(b=None), Parameter('C_0', 100))
    res = run_simulation(model, time=10, points=100, flux_map=True,
                         cleanup=True, seed=_KAPPA_SEED, verbose=False)
    simdata = res.timecourse
    ok_(len(simdata['time']) == 101)
    ok_(len(simdata['ABC']) == 101)
    ok_(simdata['time'][0] == 0)
    ok_(sorted(simdata['time'])[-1] == 10)
    fluxmap = res.flux_map
    ok_(isinstance(fluxmap, nx.MultiGraph))


@with_model
def test_kappa_wild():
    Monomer('A',['site'])
    Monomer('B',['site'])
    Initial(A(site=None), Parameter('A_0', 100))
    Initial(B(site=None), Parameter('B_0', 100))
    Initial(A(site=1) % B(site=1), Parameter('AB_0', 1000))
    Rule('deg_A', A(site=pysb.WILD) >> None, Parameter('k', 1))
    Observable('A_', A())
    run_simulation(model, time=0, seed=_KAPPA_SEED)


@raises(ValueError)
@with_model
def test_run_static_analysis_valueerror():
    Monomer('A', ['b'])
    Monomer('B', ['b'])
    Rule('A_binds_B', A(b=None) + B(b=None) >> A(b=1) % B(b=1),
         Parameter('k_A_binds_B', 1))
    Observable('AB', A(b=1) % B(b=1))
    res = run_static_analysis(model, contact_map=False, influence_map=False)



@with_model
def test_run_static_analysis_cmap():
    """Test generation of contact map by run_static_analysis"""
    Monomer('A', ['b'])
    Monomer('B', ['b'])
    Rule('A_binds_B', A(b=None) + B(b=None) >> A(b=1) % B(b=1),
         Parameter('k_A_binds_B', 1))
    Observable('AB', A(b=1) % B(b=1))
    res = run_static_analysis(model, contact_map=True, influence_map=False)
    ok_(isinstance(res.contact_map, nx.MultiGraph))
    ok_(res.influence_map is None)


@with_model
def test_run_static_analysis_imap():
    """Test generation of influence map by run_static_analysis"""
    Monomer('A', [])
    Monomer('B', ['active'], {'active': ['y', 'n']})
    Monomer('C', ['active'], {'active': ['y', 'n']})
    Initial(A(), Parameter('A_0', 100))
    Initial(B(active='n'), Parameter('B_0', 100))
    Initial(C(active='n'), Parameter('C_0', 100))
    Rule('A_activates_B',
         A() + B(active='n') >> A() + B(active='y'),
         Parameter('k_A_activates_B', 1))
    Rule('B_activates_C',
         B(active='y') + C(active='n') >> B(active='y') + C(active='y'),
         Parameter('k_B_activates_C', 1))
    res = run_static_analysis(model, contact_map=False, influence_map=True)
    ok_(isinstance(res.influence_map, nx.MultiGraph))
    ok_(res.contact_map is None)


@with_model
def test_run_static_analysis_both():
    """Test generation of both influence and contact map by run_static_analysis"""
    Monomer('A', [])
    Monomer('B', ['active'], {'active': ['y', 'n']})
    Monomer('C', ['active'], {'active': ['y', 'n']})
    Initial(A(), Parameter('A_0', 100))
    Initial(B(active='n'), Parameter('B_0', 100))
    Initial(C(active='n'), Parameter('C_0', 100))
    Rule('A_activates_B',
         A() + B(active='n') >> A() + B(active='y'),
         Parameter('k_A_activates_B', 1))
    Rule('B_activates_C',
         B(active='y') + C(active='n') >> B(active='y') + C(active='y'),
         Parameter('k_B_activates_C', 1))
    res = run_static_analysis(model, contact_map=True, influence_map=True)
    ok_(isinstance(res.influence_map, nx.MultiGraph))
    ok_(isinstance(res.contact_map, nx.MultiGraph))


@with_model
def test_contact_map():
    Monomer('A', ['b'])
    Monomer('B', ['b'])
    Rule('A_binds_B', A(b=None) + B(b=None) >> A(b=1) % B(b=1),
         Parameter('k_A_binds_B', 1))
    Observable('AB', A(b=1) % B(b=1))
    res = contact_map(model, cleanup=True)
    ok_(isinstance(res, nx.MultiGraph))


@with_model
def test_influence_map_kasa():
    Monomer('A', [])
    Monomer('B', ['active'], {'active': ['y', 'n']})
    Monomer('C', ['active'], {'active': ['y', 'n']})
    Initial(A(), Parameter('A_0', 100))
    Initial(B(active='n'), Parameter('B_0', 100))
    Initial(C(active='n'), Parameter('C_0', 100))
    Rule('A_activates_B',
         A() + B(active='n') >> A() + B(active='y'),
         Parameter('k_A_activates_B', 1))
    Rule('B_activates_C',
         B(active='y') + C(active='n') >> B(active='y') + C(active='y'),
         Parameter('k_B_activates_C', 1))
    res = influence_map(model, cleanup=True)
    ok_(isinstance(res, nx.MultiGraph))


@with_model
def test_unicode_strs():
    Monomer(u'A', [u'b'], {u'b':[u'y', u'n']})
    Rule(u'rule1', A(b=u'y') >> A(b=u'n'),
         Parameter(u'k', 1))
    Initial(A(b=u'y'), Parameter(u'A_0', 100))
    Observable(u'A_y', A(b=u'y'))
    npts = 200
    kres = run_simulation(model, time=100, points=npts, seed=_KAPPA_SEED)


@with_model
def test_none_in_rxn_pat():
    Monomer('A')
    Monomer('B')
    Rule('rule1', A() + None >> None + B(), Parameter('k', 1))
    Initial(A(), Parameter('A_0', 100))
    Observable('B_', B())
    npts = 200
    kres = run_simulation(model, time=100, points=npts, seed=_KAPPA_SEED)

    # check that rule1's reaction pattern parses with ComplexPatterns
    as_complex_pattern(A()) + None >> None + as_complex_pattern(B())


@with_model
def test_kappa_error():
    # Model with a dangling bond should raise a KasimInterfaceError
    Monomer('A', ['b'])
    Monomer('B', ['b'])
    Initial(A(b=None), Parameter('A_0', 100))

    # Can't model this as a PySB rule, since it would generate a
    # DanglingBondError. Directly inject kappa code for rule instead.
    assert_raises(KasimInterfaceError, run_simulation, model, time=10,
                  perturbation="'A_binds_B' A(b),B(b) -> A(b!1),B(b) @ "
                               "'k_A_binds_B'")


@with_model
def test_kappa_two_ghost_agents():
    Monomer('A')
    Monomer('M')
    Parameter('k', 3.0)
    Rule('synthesize_A_and_B', M() + None + None >> M() + A() + A(), k)
    Initial(M(), Parameter('M_0', 1000))
    Observable('A_', A())

    # check the ReactionPattern.__radd__ version
    rp = None + (None + A())
    assert len(rp.complex_patterns) == 3

    run_simulation(model, time=100, points=100, seed=_KAPPA_SEED)



import pysb, pysb.bng, warnings, re, sympy


class JacobianGenerator(object):

    def __init__(self, model):
        self.model = model
        self.indent_level = 0
        self.__content = None

    def get_content(self, sim_length=1):
        if self.__content == None:
            self.generate_content(sim_length=sim_length)
        assert self.indent_level == 0, "missing %d outdent call(s)" % self.indent_level
        return self.__content

    def emit(self, text=''):
        self.__content += self.indent_level * 2 * ' ' + text + "\n"

    def indent(self):
        self.indent_level += 1

    def outdent(self):
        assert self.indent_level > 0, "called outdent without matching indent"
        self.indent_level -= 1

    def generate_content(self, sim_length):
        pysb.bng.generate_equations(self.model)
        self.__content = ''
        self.emit("DECLARE")
        self.indent()
        self.generate_types()
        self.outdent()
        self.emit("END")
        self.emit()
        self.emit("MODEL %s" % self.model.name)
        self.indent()
        self.generate_parameters()
        self.generate_variables()
        self.generate_equations()
        self.outdent()
        self.emit("END")
        self.emit()
        self.emit("SIMULATION SIM")
        self.indent()
        self.generate_options()
        self.generate_units()
        self.generate_set()
        self.generate_initial()
        self.generate_schedule(sim_length)
        self.outdent()
        self.emit("END")

    def generate_types(self):
        self.emit("TYPE");
        self.indent()
        self.emit("NOTYPE = 0.0 : -1E300 : 1E300")
        self.outdent()

    def generate_parameters(self):
        self.emit("PARAMETER")
        self.indent()
        #for param_group in self.make_groups(self.model.parameters, 5):
        for param_group in self.make_groups([p for p in self.model.parameters if p.name[-2:] != '_0'], 5):
            self.emit(', '.join([p.name for p in param_group]) + ' AS REAL')
        self.outdent()

    def generate_variables(self):
        self.emit("VARIABLE")
        self.indent()
        obs_names = list(self.model.observable_groups.keys())
        for obs_group in self.make_groups(obs_names, 5):
            self.emit(', '.join(obs_group) + ' AS NOTYPE')
        var_names = ['s%d' % i for i in range(len(self.model.species))]
        for var_group in self.make_groups(var_names, 5):
            self.emit(', '.join(var_group) + ' AS NOTYPE')
        self.outdent()

    def generate_equations(self):
        self.emit("EQUATION")
        self.indent()
        obs_names = list(self.model.observable_groups.keys())
        obs_exprs = [' + '.join('%g * s%s' % g for g in self.model.observable_groups[name]) for name in obs_names]
        for obs in zip(obs_names, obs_exprs):
            self.emit('%s = %s;' % obs)
        var_names = ['s%d' % i for i in range(len(self.model.species))]
        for (name, ode) in zip(var_names, self.model.odes):
            expression = sympy.sstr(ode)
            expression = re.sub(r'\*\*', '^', expression)
            self.emit('$%s = %s;' % (name, expression))
        self.outdent()

    def generate_options(self):
        self.emit("OPTIONS")
        self.indent()
        self.outdent()

    def generate_units(self):
        self.emit("UNIT")
        self.indent()
        self.emit("M AS " + self.model.name)
        self.outdent()

    def generate_set(self):
        self.emit("SET")
        self.indent()
        for p in self.model.parameters:
            #self.emit("M.%s := %g;" % (p.name, p.value))
            if p.name[-2:] != '_0':
                self.emit("M.%s := %g;" % (p.name, p.value))
        self.outdent()

    def generate_initial(self):
        self.emit("INITIAL")
        self.indent()
        vars_unseen = set(['s%d' % i for i in range(len(self.model.species))])
        for (pattern, param) in self.model.initial_conditions:
            sname = 's%d' % self.model.get_species_index(pattern)
            vars_unseen.remove(sname)
            #self.emit("M.%s = M.%s;" % (sname, param.name))
            self.emit("M.%s = %g;" % (sname, param.value))
        for sname in sorted(vars_unseen):
            self.emit("M.%s = 0;" % sname)
        self.outdent()

    def generate_schedule(self, sim_length):
        self.emit("SCHEDULE")
        self.indent()
        self.emit("CONTINUE FOR %d" % sim_length);
        self.outdent()

    def make_groups(self, elements, size):
        groups = []
        offsets = list(range(0, len(elements), size)) + [None]
        for i in range(0, len(offsets) - 1):
            groups.append(elements[offsets[i]:offsets[i+1]])
        return groups



"""
Wrapper functions for running the Kappa programs *KaSim* and *KaSa*.

The path to the directory containing the KaSim and KaSa executables can be
specified in one of three ways:

- set the KAPPAPATH environment variable to the KaSim directory
- move Kappa to /usr/local/share/KaSim (macOS, Linux) or
  c:\Program Files\KaSim (Windows)
- set the path using the :py:func:`pysb.pathfinder.set_path` function at
  runtime
"""

from __future__ import print_function as _
import pysb
import pysb.pathfinder as pf
from pysb.generator.kappa import KappaGenerator
import os
import subprocess
import re
import numpy as np
import tempfile
import shutil
import warnings
from collections import namedtuple
from pysb.util import read_dot

try:
    from future_builtins import zip
except ImportError:
    pass


def set_kappa_path(path):
    """Set the path to the KaSim and KaSa executables.

    Deprecated. Use pysb.pathfinder.set_path() instead.

    Parameters
    ----------
    path: string
        Directory containing KaSim and KaSa executables.
    """
    warnings.warn("Function %s() is deprecated; use "
                  "pysb.pathfinder.set_path() instead" %
                  set_kappa_path.__name__, category=DeprecationWarning,
                  stacklevel=2)
    pf.set_path('kasim', path)
    pf.set_path('kasa', path)


class KasimInterfaceError(RuntimeError):
    pass


class KasaInterfaceError(RuntimeError):
    pass

StaticAnalysisResult = namedtuple('StaticAnalysisResult',
                                  ['contact_map', 'influence_map'])

SimulationResult = namedtuple('SimulationResult',
                                  ['timecourse', 'flux_map'])


def run_simulation(model, time=10000, points=200, cleanup=True,
                   output_prefix=None, output_dir=None, flux_map=False,
                   perturbation=None, seed=None, verbose=False):
    """Runs the given model using KaSim and returns the parsed results.

    Parameters
    ----------
    model : pysb.core.Model
        The model to simulate/analyze using KaSim.
    time : number
        The amount of time (in arbitrary units) to run a simulation.
        Identical to the -u time -l argument when using KaSim at the command
        line.
        Default value is 10000. If set to 0, no simulation will be run.
    points : integer
        The number of data points to collect for plotting.
        Note that this is not identical to the -p argument of KaSim when
        called from the command line, which denotes plot period (time interval
        between points in plot).
        Default value is 200. Note that the number of points actually returned
        by the simulator will be points + 1 (including the 0 point).
    cleanup : boolean
        Specifies whether output files produced by KaSim should be deleted
        after execution is completed. Default value is True.
    output_prefix: str
        Prefix of the temporary directory name. Default is
        'tmpKappa_<model name>_'.
    output_dir : string
        The directory in which to create the temporary directory for
        the .ka and other output files. Defaults to the system temporary file
        directory (e.g. /tmp). If the specified directory does not exist,
        an Exception is thrown.
    flux_map: boolean
        Specifies whether or not to produce the flux map (generated over the
        full duration of the simulation). Default value is False.
    perturbation : string or None
        Optional perturbation language syntax to be appended to the Kappa file.
        See KaSim manual for more details. Default value is None (no
        perturbation).
    seed : integer
        A seed integer for KaSim random number generator. Set to None to
        allow KaSim to use a random seed (default) or supply a seed for
        deterministic behaviour (e.g. for testing)
    verbose : boolean
        Whether to pass the output of KaSim through to stdout/stderr.

    Returns
    -------
    If flux_map is False, returns the kasim simulation data as a Numpy ndarray.
    Data is accessed using the syntax::

            results[index_name]

    The index 'time' gives the time coordinates of the simulation. Data for the
    observables can be accessed by indexing the array with the names of the
    observables. Each entry in the ndarray has length points + 1, due to the
    inclusion of both the zero point and the final timepoint.

    If flux_map is True, returns an instance of SimulationResult, a namedtuple
    with two members, `timecourse` and `flux_map`. The `timecourse` field
    contains the simulation ndarray, and the `flux_map` field is an instance of
    a networkx MultiGraph containing the flux map. For details on viewing
    the flux map graphically see :func:`run_static_analysis` (notes section).
    """

    gen = KappaGenerator(model)

    if output_prefix is None:
        output_prefix = 'tmpKappa_%s_' % model.name

    base_directory = tempfile.mkdtemp(prefix=output_prefix, dir=output_dir)

    base_filename = os.path.join(base_directory, model.name)
    kappa_filename = base_filename + '.ka'
    fm_filename = base_filename + '_fm.dot'
    out_filename = base_filename + '.out'

    if points == 0:
        raise ValueError('The number of data points cannot be zero.')
    plot_period = (float(time) / points) if time > 0 else 1.0

    args = ['-i', kappa_filename, '-u', 'time', '-l', str(time),
            '-p', '%.5f' % plot_period, '-o', out_filename]

    if seed:
        args.extend(['-seed', str(seed)])

    # Generate the Kappa model code from the PySB model and write it to
    # the Kappa file:
    with open(kappa_filename, 'w') as kappa_file:
        kappa_file.write(gen.get_content())
        # If desired, add instructions to the kappa file to generate the
        # flux map:
        if flux_map:
            kappa_file.write('%%mod: [true] do $DIN "%s" [true];\n' %
                             fm_filename)

        # If any perturbation language code has been passed in, add it to
        # the Kappa file:
        if perturbation:
            kappa_file.write('\n%s\n' % perturbation)

    # Run KaSim
    kasim_path = pf.get_path('kasim')
    p = subprocess.Popen([kasim_path] + args,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         cwd=base_directory)
    if verbose:
        for line in iter(p.stdout.readline, b''):
            print('@@', line, end='')
    (p_out, p_err) = p.communicate()

    if p.returncode:
        raise KasimInterfaceError(
            p_out.decode('utf8') + '\n' + p_err.decode('utf8'))

    # The simulation data, as a numpy array
    data = _parse_kasim_outfile(out_filename)

    if flux_map:
        try:
            flux_graph = read_dot(fm_filename)
        except ImportError:
            if cleanup:
                raise
            else:
                warnings.warn(
                        "The pydot library could not be "
                        "imported, so no MultiGraph "
                        "object returned (returning None); flux map "
                        "dot file available at %s" % fm_filename)
                flux_graph = None

    if cleanup:
        shutil.rmtree(base_directory)

    # If a flux map was generated, return both the simulation output and the
    # flux map as a networkx multigraph
    if flux_map:
        return SimulationResult(data, flux_graph)
    # If no flux map was requested, return only the simulation data
    else:
        return data


def run_static_analysis(model, influence_map=False, contact_map=False,
                        cleanup=True, output_prefix=None, output_dir=None,
                        verbose=False):
    """Run static analysis (KaSa) on to get the contact and influence maps.

    If neither influence_map nor contact_map are set to True, then a ValueError
    is raised.

    Parameters
    ----------
    model : pysb.core.Model
        The model to simulate/analyze using KaSa.
    influence_map : boolean
        Whether to compute the influence map.
    contact_map : boolean
        Whether to compute the contact map.
    cleanup : boolean
        Specifies whether output files produced by KaSa should be deleted
        after execution is completed. Default value is True.
    output_prefix: str
        Prefix of the temporary directory name. Default is
        'tmpKappa_<model name>_'.
    output_dir : string
        The directory in which to create the temporary directory for
        the .ka and other output files. Defaults to the system temporary file
        directory (e.g. /tmp). If the specified directory does not exist,
        an Exception is thrown.
    verbose : boolean
        Whether to pass the output of KaSa through to stdout/stderr.

    Returns
    -------
    StaticAnalysisResult, a namedtuple with two fields, `contact_map` and
    `influence_map`, each containing the respective result as an instance
    of a networkx MultiGraph. If the either the contact_map or influence_map
    argument to the function is False, the corresponding entry in the
    StaticAnalysisResult returned by the function will be None.

    Notes
    -----
    To view a networkx file graphically, use `draw_network`::

        import networkx as nx
        nx.draw_networkx(g, with_labels=True)

    You can use `graphviz_layout` to use graphviz for layout (requires pydot
    library)::

        import networkx as nx
        pos = nx.drawing.nx_pydot.graphviz_layout(g, prog='dot')
        nx.draw_networkx(g, pos, with_labels=True)

    For further information, see the networkx documentation on visualization:
    https://networkx.github.io/documentation/latest/reference/drawing.html
    """

    # Make sure the user has asked for an output!
    if not influence_map and not contact_map:
        raise ValueError('Either contact_map or influence_map (or both) must '
                         'be set to True in order to perform static analysis.')

    gen = KappaGenerator(model, _warn_no_ic=False)

    if output_prefix is None:
        output_prefix = 'tmpKappa_%s_' % model.name

    base_directory = tempfile.mkdtemp(prefix=output_prefix, dir=output_dir)

    base_filename = os.path.join(base_directory, str(model.name))
    kappa_filename = base_filename + '.ka'
    im_filename = base_filename + '_im.dot'
    cm_filename = base_filename + '_cm.dot'

    # NOTE: in the args passed to KaSa, the directory for the .dot files is
    # specified by the --output_directory option, and the output_contact_map
    # and output_influence_map should only be the base filenames (without
    # a directory prefix).
    # Contact map args:
    if contact_map:
        cm_args = ['--compute-contact-map', '--output-contact-map',
                   os.path.basename(cm_filename),
                   '--output-contact-map-directory', base_directory]
    else:
        cm_args = ['--no-compute-contact-map']
    # Influence map args:
    if influence_map:
        im_args = ['--compute-influence-map', '--output-influence-map',
                   os.path.basename(im_filename),
                   '--output-influence-map-directory', base_directory]
    else:
        im_args = ['--no-compute-influence-map']
    # Full arg list
    args = [kappa_filename] + cm_args + im_args

    # Generate the Kappa model code from the PySB model and write it to
    # the Kappa file:
    with open(kappa_filename, 'w') as kappa_file:
        kappa_file.write(gen.get_content())

    # Run KaSa using the given args
    kasa_path = pf.get_path('kasa')
    p = subprocess.Popen([kasa_path] + args,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         cwd=base_directory)
    if verbose:
        for line in iter(p.stdout.readline, b''):
            print('@@', line, end='')
    (p_out, p_err) = p.communicate()

    if p.returncode:
        raise KasaInterfaceError(
            p_out.decode('utf8') + '\n' + p_err.decode('utf8'))

    # Try to create the graphviz objects from the .dot files created
    try:
        # Convert the contact map to a Graph
        cmap = read_dot(cm_filename) if contact_map else None
        imap = read_dot(im_filename) if influence_map else None
    except ImportError:
        if cleanup:
            raise
        else:
            warnings.warn(
                    "The pydot library could not be "
                    "imported, so no MultiGraph "
                    "object returned (returning None); "
                    "contact/influence maps available at %s" %
                    base_directory)
            cmap = None
            imap = None

    # Clean up the temp directory if desired
    if cleanup:
        shutil.rmtree(base_directory)

    return StaticAnalysisResult(cmap, imap)


def contact_map(model, **kwargs):
    """Generates the contact map via KaSa.

    Parameters
    ----------
    model : pysb.core.Model
        The model for generating the influence map.
    **kwargs : other keyword arguments
        Any other keyword arguments are passed to the function
        :py:func:`run_static_analysis`.

    Returns
    -------
    networkx MultiGraph object containing the contact map. For details on
    viewing the contact map graphically see :func:`run_static_analysis` (notes
    section).
    """
    kasa_result = run_static_analysis(model, influence_map=False,
                                    contact_map=True, **kwargs)
    return kasa_result.contact_map


def influence_map(model, **kwargs):
    """Generates the influence map via KaSa.

    Parameters
    ----------
    model : pysb.core.Model
        The model for generating the influence map.
    **kwargs : other keyword arguments
        Any other keyword arguments are passed to the function
        :py:func:`run_static_analysis`.

    Returns
    -------
    networkx MultiGraph object containing the influence map. For details on
    viewing the influence map graphically see :func:`run_static_analysis`
    (notes section).
    """

    kasa_result = run_static_analysis(model, influence_map=True,
                                    contact_map=False, **kwargs)
    return kasa_result.influence_map

### "PRIVATE" Functions ###############################################

def _parse_kasim_outfile(out_filename):
    """
    Parses the KaSim .out file into a Numpy ndarray.

    Parameters
    ----------
    out_filename : string
        String specifying the location of the .out filename produced by KaSim.

    Returns
    -------
    numpy.ndarray
        Returns the KaSim simulation data as a Numpy ndarray. Data is accessed
        using the syntax::

            results[index_name]

        The index 'time' gives the data for the time coordinates of the
        simulation. Data for the observables can be accessed by indexing the
        array with the names of the observables.
    """
    try:
        with open(out_filename, 'r') as fh:
            for header_line in fh:
                if header_line[0] != '#':
                    break

            # Load the output file as a numpy record array, skip the name row
            arr = np.loadtxt(fh, dtype=float, delimiter=',')

        raw_names = [term.strip() for term in re.split(',', header_line)]
        column_names = []

        # Get rid of the quotes surrounding the observable names
        for raw_name in raw_names:
            mo = re.match('"(.*)"', raw_name)
            if mo:
                name = mo.group(1)
                # Rename the time column to remain backwards compatible
                if name == '[T]':
                    name = 'time'
                column_names.append(name)
            else:
                column_names.append(raw_name)

        # Create the dtype argument for the numpy record array
        dt = list(zip(column_names, ('float', ) * len(column_names)))

        recarr = arr.view(dt)
    except Exception as e:
        raise Exception("problem parsing KaSim outfile: " + str(e))

    return recarr



from pysb.simulator.base import Simulator, SimulationResult, SimulatorException
from pysb.bng import generate_equations
from pysb.export.stochkit import StochKitExporter
import numpy as np
import subprocess
import os
import shutil
import tempfile
from pysb.pathfinder import get_path
from pysb.logging import EXTENDED_DEBUG


class StochKitSimulator(Simulator):
    """
    Interface to the StochKit 2 stochastic simulation toolkit

    StochKit can be installed from GitHub:
    https://github.com/stochss/stochkit

    This class is inspired by the `gillespy
    <https://github.com/JohnAbel/gillespy>` library, but has been optimised
    for use with PySB.

    .. warning::
        The interface for this class is considered experimental and may
        change without warning as PySB is updated.

    Parameters
    ----------
    model : pysb.Model
        Model to simulate.
    tspan : vector-like, optional
        Time values over which to simulate. The first and last values define
        the time range. Returned trajectories are sampled at every value unless
        the simulation is interrupted for some reason, e.g., due to
        satisfaction of a logical stopping criterion (see 'tout' below).
    initials : vector-like or dict, optional
        Values to use for the initial condition of all species. Ordering is
        determined by the order of model.species. If not specified, initial
        conditions will be taken from model.initial_conditions (with
        initial condition parameter values taken from `param_values` if
        specified).
    param_values : vector-like or dict, optional
        Values to use for every parameter in the model. Ordering is
        determined by the order of model.parameters.
        If passed as a dictionary, keys must be parameter names.
        If not specified, parameter values will be taken directly from
        model.parameters.
    verbose : bool or int, optional (default: False)
        Sets the verbosity level of the logger. See the logging levels and
        constants from Python's logging module for interpretation of integer
        values. False is equal to the PySB default level (currently WARNING),
        True is equal to DEBUG.
    **kwargs : dict
        Extra keyword arguments, including:

        * ``cleanup``: Boolean, delete directory after completion if True

    Examples
    --------
    Simulate a model and display the results for an observable:

    >>> from pysb.examples.robertson import model
    >>> import numpy as np
    >>> np.set_printoptions(precision=4)
    >>> sim = StochKitSimulator(model, tspan=np.linspace(0, 10, 5))

    Here we supply a "seed" to the random number generator for deterministic
    results, but for most purposes it is recommended to leave this blank.

    >>> simulation_result = sim.run(n_runs=2, seed=123456)

    A_total trajectory for first run

    >>> print(simulation_result.observables[0]['A_total']) \
        #doctest: +NORMALIZE_WHITESPACE
    [1.  0.  0.  0.  0.]

    A_total trajectory for second run

    >>> print(simulation_result.observables[1]['A_total']) \
        #doctest: +SKIP
    [1.  1.  1.  0.  0.]

    For further information on retrieving trajectories (species,
    observables, expressions over time) from the ``simulation_result``
    object returned by :func:`run`, see the examples under the
    :class:`SimulationResult` class.
    """
    _supports = {'multi_initials': True, 'multi_param_values': True}

    def __init__(self, model, tspan=None, initials=None,
                 param_values=None, verbose=False, **kwargs):
        super(StochKitSimulator, self).__init__(model,
                                                tspan=tspan,
                                                initials=initials,
                                                param_values=param_values,
                                                verbose=verbose,
                                                **kwargs)
        self.cleanup = kwargs.get('cleanup', True)
        self._outdir = None
        generate_equations(self._model,
                           cleanup=self.cleanup,
                           verbose=self.verbose)

    def _run_stochkit(self, t=20, t_length=100, number_of_trajectories=1,
                      seed=None, algorithm='ssa', method=None,
                      num_processors=1, stats=False, epsilon=None,
                      threshold=None):

        extra_args = '-p {:d}'.format(num_processors)

        # Random seed for stochastic simulation
        if seed is not None:
            extra_args += ' --seed {:d}'.format(seed)

        # Keep all the trajectories by default
        extra_args += ' --keep-trajectories'

        # Number of trajectories
        extra_args += ' --realizations {:d}'.format(number_of_trajectories)

        # We generally don't need the extra stats
        if not stats:
            extra_args += ' --no-stats'

        if method is not None:  # This only works for StochKit 2.1
            extra_args += ' --method {}'.format(method)

        if epsilon is not None:
            extra_args += ' --epsilon {:f}'.format(epsilon)

        if threshold is not None:
            extra_args += ' --threshold {:d}'.format(threshold)

        # Find binary for selected algorithm (SSA, Tau-leaping, ...)
        if algorithm not in ['ssa', 'tau_leaping']:
            raise SimulatorException(
                "algorithm must be 'ssa' or 'tau_leaping'")

        executable = get_path('stochkit_{}'.format(algorithm))

        # Output model file to directory
        fname = os.path.join(self._outdir, 'pysb.xml')

        trajectories = []
        for i in range(len(self.initials)):
            # We write all StochKit output files to a temporary folder
            prefix_outdir = os.path.join(self._outdir, 'output_{}'.format(i))

            # Export model file
            stoch_xml = StochKitExporter(self._model).export(
                self.initials[i], self.param_values[i])
            self._logger.log(EXTENDED_DEBUG, 'StochKit XML:\n%s' % stoch_xml)
            with open(fname, 'wt') as f:
                f.write(stoch_xml)

            # Assemble the argument list
            args = '--model {} --out-dir {} -t {:f} -i {:d}'.format(
                fname, prefix_outdir, t, t_length - 1)

            # If we are using local mode, shell out and run StochKit
            # (SSA or Tau-leaping or ODE)
            cmd = '{} {} {}'.format(executable, args, extra_args)
            self._logger.debug("StochKit run {} of {} (cmd: {})".format(
                (i + 1), len(self.initials), cmd))

            # Execute
            try:
                handle = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                          stderr=subprocess.PIPE, shell=True)
                return_code = handle.wait()
            except OSError as e:
                raise SimulatorException("StochKit execution failed: \
                {0}\n{1}".format(cmd, e))

            try:
                stderr = handle.stderr.read()
            except Exception as e:
                stderr = 'Error reading stderr: {0}'.format(e)
            try:
                stdout = handle.stdout.read()
            except Exception as e:
                stdout = 'Error reading stdout: {0}'.format(e)

            if return_code != 0:
                raise SimulatorException("Solver execution failed: \
                '{0}' output:\nSTDOUT:\n{1}\nSTDERR:\n{2}".format(
                    cmd, stdout, stderr))

            traj_dir = os.path.join(prefix_outdir, 'trajectories')
            try:
                trajectories.extend([np.loadtxt(os.path.join(
                    traj_dir, f)) for f in sorted(os.listdir(traj_dir))])
            except Exception as e:
                raise SimulatorException(
                    "Error reading StochKit trajectories: {0}"
                    "\nSTDOUT:{1}\nSTDERR:{2}".format(e, stdout, stderr))

            if len(trajectories) == 0 or len(stderr) != 0:
                raise SimulatorException("Solver execution failed: \
                '{0}' output:\nSTDOUT:\n{1}\nSTDERR:\n{2}".format(
                    cmd, stdout, stderr))

            self._logger.debug("StochKit STDOUT:\n{0}".format(stdout))

        # Return data
        return trajectories

    def run(self, tspan=None, initials=None, param_values=None, n_runs=1,
            algorithm='ssa', output_dir=None,
            num_processors=1, seed=None, method=None, stats=False,
            epsilon=None, threshold=None):
        """
        Run a simulation and returns the result (trajectories)

        .. note::
            In early versions of the Simulator class, ``tspan``, ``initials``
            and ``param_values`` supplied to this method persisted to future
            :func:`run` calls. This is no longer the case.

        Parameters
        ----------
        tspan
        initials
        param_values
            See parameter definitions in :class:`StochKitSimulator`.
        n_runs : int
            The number of simulation runs per parameter set. The total
            number of simulations is therefore n_runs * max(len(initials),
            len(param_values))
        algorithm : str
            Choice of 'ssa' (Gillespie's stochastic simulation algorithm) or
            'tau_leaping' (Tau leaping algorithm)
        output_dir : str or None
            Directory for StochKit output, or None for a system-specific
            temporary directory
        num_processors : int
            Number of CPU cores for StochKit to use (default: 1)
        seed : int or None
            A random number seed for StochKit. Set to any integer value for
            deterministic behavior.
        method : str or None
            StochKit "method" argument, default None. Only used by StochKit
            2.1 (not yet released at time of writing).
        stats : bool
            Ask StochKit to generate simulation summary statistics if True
        epsilon : float or None
            Tolerance parameter for tau-leaping algorithm
        threshold : int or None
            Threshold parameter for tau-leaping algorithm

        Returns
        -------
        A :class:`SimulationResult` object
        """
        super(StochKitSimulator, self).run(tspan=tspan,
                                           initials=initials,
                                           param_values=param_values,
                                           _run_kwargs=locals())

        self._logger.info('Running StochKit with {:d} parameter sets, '
                          '{:d} repeats ({:d} simulations total)'.format(
                           len(self.initials), n_runs, len(self.initials) *
                           n_runs))

        if output_dir is None:
            self._outdir = tempfile.mkdtemp()
        else:
            self._outdir = output_dir

        # Calculate time intervals and validate
        t_range = self.tspan[-1] - self.tspan[0]
        t_length = len(self.tspan)
        if not np.allclose(self.tspan, np.linspace(0, self.tspan[-1],
                                                   t_length)):
            raise SimulatorException('StochKit requires tspan to be linearly '
                                     'spaced starting at t=0')

        try:
            trajectories = self._run_stochkit(t=t_range,
                                              number_of_trajectories=n_runs,
                                              t_length=t_length,
                                              seed=seed,
                                              algorithm=algorithm,
                                              method=method,
                                              num_processors=num_processors,
                                              stats=stats,
                                              epsilon=epsilon,
                                              threshold=threshold)
        finally:
            if self.cleanup:
                try:
                    shutil.rmtree(self._outdir)
                except OSError:
                    pass

        # set output time points
        trajectories_array = np.array(trajectories)
        self.tout = trajectories_array[:, :, 0] + self.tspan[0]
        # species
        species = trajectories_array[:, :, 1:]
        return SimulationResult(self, self.tout, species,
                                simulations_per_param_set=n_runs)



#!/usr/bin/env python

from __future__ import print_function
from pysb.simulator import ScipyOdeSimulator
from matplotlib.pyplot import plot, legend, show
from matplotlib.pylab import linspace
from kinase_cascade import model


tspan = linspace(0, 1200)
print("Simulating...")
yfull = ScipyOdeSimulator(model).run(tspan=tspan).all
plot(tspan, yfull['ppMEK'], label='ppMEK')
plot(tspan, yfull['ppERK'], label='ppERK')
legend(loc='upper left')
show()



from pysb.testing import *
from pysb.core import *
from functools import partial
import itertools


@with_model
def test_component_names_valid():
    for name in 'a', 'B', 'AbC', 'dEf', '_', '_7', '__a01b__999x_x___':
        c = Monomer(name, _export=False)
        eq_(c.name, name)
        # Before the component is added, we should not be able to find it
        assert not hasattr(model.components, name)
        # Add the element to a model and try to access it as attribute and item
        model.add_component(c)
        assert_equal(model.components[name], c)
        assert_equal(getattr(model.components, name), c)


@with_model
def test_component_name_existing_attribute():
    for name in ('_map', 'keys'):
        c = Monomer(name, _export=False)
        model.add_component(c)
        # When using an existing attribute name like_map, we should get able to
        # get it as an item, but not as an attribute
        assert_equal(model.components[name], c)
        assert_not_equal(getattr(model.components, name), c)


def test_component_names_invalid():
    for name in 'a!', '!B', 'A!bC~`\\', '_!', '_!7', '__a01b  999x_x___!':
        assert_raises(InvalidComponentNameError, Component, name, _export=False)

def test_monomer():
    sites = ['x', 'y', 'z']
    states = {'y': ['foo', 'bar', 'baz'], 'x': ['e']}
    m = Monomer('A', sites, states, _export=False)
    assert_equal(m.sites, sites)
    assert_equal(m.site_states, states)
    assert_equal(type(m()), MonomerPattern)

    assert_raises(ValueError, Monomer, 'A', 'x', _export=False)
    assert_raises(Exception, Monomer, 'A', 'x', 'x', _export=False)
    assert_raises(Exception, Monomer, 'A', ['x'], {'y': ['a']}, _export=False)
    assert_raises(Exception, Monomer, 'A', ['x'], {'x': [1]}, _export=False)

@with_model
def test_monomer_model():
    Monomer('A')
    ok_(A in model.monomers)
    ok_(A in model.all_components())
    ok_(A not in model.all_components() - model.monomers)


@with_model
def test_monomer_rename_self_exporter():
    Monomer('A').rename('B')
    assert 'B' in model.monomers.keys()
    assert 'A' not in model.monomers.keys()


@with_model
def test_monomer_rename_non_self_exported_component():
    mon = Monomer('A', _export=False)
    mon.rename('B')
    assert mon.name == 'B'
    model.add_component(mon)
    mon.rename('C')
    assert mon.name == 'C'
    assert model.monomers['C'] == mon
    assert model.monomers.keys() == ['C']


def test_monomer_rename_non_self_exported_model():
    model = Model(_export=False)
    mon = Monomer('A', _export=False)
    model.add_component(mon)
    mon.rename('B')
    assert model.monomers.keys() == ['B']
    assert model.monomers['B'] == mon


@with_model
def test_invalid_state():
    Monomer('A', ['a', 'b'], {'a': ['a1', 'a2'], 'b': ['b1']})
    # Specify invalid state in Monomer.__call__
    assert_raises(ValueError, A, a='spam')
    # Specify invalid state in MonomerPattern.__call__
    assert_raises(ValueError, A(a='a1'), b='spam')


@with_model
def test_initial():
    Monomer('A', ['s'])
    Parameter('A_0')
    Initial(A(s=None), A_0)
    assert_raises_iice = partial(assert_raises, InvalidInitialConditionError,
                                 Initial)
    assert_raises_iice('not a complexpattern', A_0)
    assert_raises_iice(A(), A_0)
    assert_raises_iice(A(s=None), A_0)
    assert_raises_iice(MatchOnce(A(s=None)), A_0)

@with_model
def test_model_pickle():
    import pickle
    A = Monomer('A', _export=False)
    B = Monomer('B', ['x', 'y'], {'x': ['e', 'f']}, _export=False)
    k = Parameter('k', 1.0, _export=False)
    r = Rule('r', A() + B(x='e', y=WILD) >> A() % B(x='f', y=None), k,
             _export=False)
    o = Observable('o', A() % B(), _export=False)
    e = Expression('e', k * o, _export=False)
    c = Compartment('c', None, 3, k, _export=False)
    for comp in [A, B, k, r, o, e, c]:
        model.add_component(comp)
    model.add_component(c)
    Initial(A() ** c, k)
    assert_equal(len(model.all_components()), 7)
    model2 = pickle.loads(pickle.dumps(model))
    check_model_against_component_list(model, model2.all_components())

@with_model
def test_monomer_as_reaction_pattern():
    A = Monomer('A', _export=False)
    as_reaction_pattern(A)

@with_model
def test_monomer_as_complex_pattern():
    A = Monomer('A', _export=False)
    as_complex_pattern(A)

@with_model
def test_monomerpattern():
    A = Monomer('A',sites=['a'],site_states={'a':['u','p']},_export=False)
    Aw = A(a=('u', ANY))


@with_model
def test_duplicate_monomer_error():
    A = Monomer('A', ['a'])
    assert_raises(DuplicateMonomerError, (A(a=1) % A(a=1)), a=2)

@with_model
def test_observable_constructor_with_monomer():
    A = Monomer('A', _export=False)
    o = Observable('o', A, _export=False)

@with_model
def test_compartment_initial_error():
    Monomer('A', ['s'])
    Parameter('A_0', 2.0)
    c1 = Compartment("C1")
    c2 = Compartment("C2")
    Initial(A(s=None)**c1, A_0)
    Initial(A(s=None)**c2, A_0)

@with_model
def test_monomer_pattern_add_to_none():
    """Ensure that MonomerPattern + None returns a ReactionPattern."""
    Monomer('A', ['s'])
    ok_(isinstance(A() + None, ReactionPattern),
        'A() + None did not return a ReactionPattern.')

@with_model
def test_complex_pattern_add_to_none():
    """Ensure that ComplexPattern + None returns a ReactionPattern."""
    Monomer('A', ['s'])
    ok_(isinstance(A(s=1) % A(s=1) + None, ReactionPattern),
        'A(s=1) % A(s=1) + None did not return a ReactionPattern.')

@with_model
def test_reaction_pattern_add_to_none():
    """Ensure that ReactionPattern + None returns a ReactionPattern."""
    Monomer('A', ['s'])
    cp = A(s=1) % A(s=1)
    rp = cp + cp
    ok_(isinstance(rp + None, ReactionPattern),
        'ReactionPattern + None did not return a ReactionPattern.')

@with_model
def test_complex_pattern_call():
    """Ensure ComplexPattern calling (refinement) works as expected."""
    Monomer('A', ['w', 'x'], {'x': ('e', 'f')})
    Monomer('B', ['y', 'z'], {'z': ('g', 'h')})
    cp = A(w=1, x='e') % B(y=1, z='g')
    r = {'x': 'f', 'z': 'h'} # refinement for complexpattern
    ok_(cp(**r))
    ok_(cp(**r).monomer_patterns[0].site_conditions['x'] == r['x'])
    ok_(cp(r))
    ok_(cp(r).monomer_patterns[0].site_conditions['x'] == r['x'])
    assert_raises(RedundantSiteConditionsError, cp, {'x': 'f'}, z='h')

@with_model
def test_monomer_unicode():
    Monomer(u'A', [u's'], {u's': [u's1', u's2']})


def _check_pattern_equivalence(complex_pattern_list, equivalent=True):
    """ Check all complex pattern permutations are equivalent (or not) """
    for cp0, cp1 in itertools.permutations(complex_pattern_list, 2):
        assert cp0.is_equivalent_to(cp1) == equivalent


@with_model
def test_complex_pattern_equivalence_compartments():
    Monomer('A')
    Monomer('B')

    Compartment('C1')
    Compartment('C2')

    cp0 = ComplexPattern([A()], compartment=C1)  # Only species compartment
    cp1 = as_complex_pattern(A() ** C1)          # Only monomer compartment
    cp2 = ComplexPattern([A() ** C1], compartment=C1)  # Both compartments

    _check_pattern_equivalence((cp0, cp1, cp2))

    cp3 = (A() % B()) ** C1
    cp4 = A() ** C1 % B() ** C1
    cp5 = (A() ** C1 % B()) ** C1
    # Species compartment is ignored if all monomer patterns have a compartment
    cp6 = (A() ** C1 % B() ** C1) ** C2

    _check_pattern_equivalence((cp3, cp4, cp5, cp6))

    # Species compartment should not override a specified monomer compartment
    cp7 = (A() ** C1 % B() ** C2) ** C1
    _check_pattern_equivalence((cp5, cp7), equivalent=False)


@with_model
def test_complex_pattern_equivalence_state():
    """Ensure CP equivalence handles site states."""
    Monomer('A', ['s', 't'], {'t': ['x', 'y', 'z']})
    cp0 = A(s=1, t='x') % A(s=1, t='y')
    cp1 = A(s=1, t='y') % A(s=1, t='x')
    cp2 = A(s=1, t='x') % A(s=1, t='z')
    cp3 = A(s=1, t='z') % A(s=1, t='y')
    _check_pattern_equivalence((cp0, cp1))
    _check_pattern_equivalence((cp0, cp2), False)
    _check_pattern_equivalence((cp0, cp3), False)


@with_model
def test_complex_pattern_equivalence_bond_state():
    """Ensure CP equivalence handles bond and state on the same site."""
    Monomer('A', ['s'], {'s': ['x', 'y', 'z']})
    cp0 = A(s=('x', 1)) % A(s=('y', 1))
    cp1 = A(s=('y', 1)) % A(s=('x', 1))
    cp2 = A(s=('z', 1)) % A(s=('y', 1))
    cp3 = A(s='x') % A(s='y')
    _check_pattern_equivalence((cp0, cp1))
    _check_pattern_equivalence((cp0, cp2), False)
    _check_pattern_equivalence((cp0, cp3), False)


@with_model
def test_complex_pattern_equivalence_bond_numbering():
    """Ensure CP equivalence is insensitive to bond numbers."""
    Monomer('A', ['s'])
    cp0 = A(s=1) % A(s=1)
    cp1 = A(s=2) % A(s=2)
    _check_pattern_equivalence((cp0, cp1))


@with_model
def test_complex_pattern_equivalence_monomer_pattern_ordering():
    """Ensure CP equivalence is insensitive to MP order."""
    Monomer('A', ['s1', 's2'])
    cp0 = A(s1=1, s2=2) % A(s1=2, s2=1)
    cp1 = A(s1=2, s2=1) % A(s1=1, s2=2)
    _check_pattern_equivalence((cp0, cp1))


@with_model
def test_complex_pattern_equivalence_compartments():
    """Ensure CP equivalence is insensitive to Compartment"""
    Monomer('A', ['s1'])
    Compartment('C')
    cp0 = (A(s1=1) % A(s1=1)) ** C
    cp1 = (A(s1=1) ** C) % (A(s1=1) ** C)
    _check_pattern_equivalence((cp0, cp1))


@with_model
def test_reaction_pattern_match_complex_pattern_ordering():
    """Ensure CP equivalence is insensitive to MP order."""
    Monomer('A', ['s1', 's2'])
    cp0 = A(s1=1, s2=2) % A(s1=2, s2=1)
    cp1 = A(s1=2, s2=1) % A(s1=1, s2=2)
    rp0 = cp0 + cp1
    rp1 = cp1 + cp0
    rp2 = cp0 + cp0
    assert rp0.matches(rp1)
    assert rp1.matches(rp0)
    assert rp2.matches(rp0)


@with_model
def test_concreteness():
    Monomer('A', ['s'], {'s': ['x']})
    assert not (A(s=1) % A(s=1)).is_concrete()
    assert not (A(s=('x', 1)) % A(s=1)).is_concrete()
    assert (A(s=('x', 1)) % A(s=('x', 1))).is_concrete()
    assert (A(s='x')).is_concrete()
    assert not A().is_concrete()
    assert not A(s=('x', ANY)).is_concrete()
    assert not A(s=WILD).is_concrete()

    Monomer('B', ['s'])
    assert not B().is_concrete()
    assert not B(s=ANY).is_concrete()
    assert B(s=1).is_concrete()

    Monomer('C')
    assert C().is_concrete()

    # Tests with compartments
    Compartment('cell')
    assert not C().is_concrete()
    assert (C() ** cell).is_concrete()


@with_model
def test_dangling_bond():
    Monomer('A', ['a'])
    Parameter('kf', 1.0)
    assert_raises(DanglingBondError, as_reaction_pattern, A(a=1) % A(a=None))



#!/usr/bin/env python
""" Run the Extrinsic Apoptosis Reaction Model (EARM) using StochKit's
stochastic simulation algorithm (SSA) implementation.
"""

import matplotlib.pyplot as plt
import numpy as np

from earm_1_0 import model
from pysb.simulator import StochKitSimulator


def plot_mean_min_max(name, title=None):
    x = np.array([tr[:][name] for tr in trajectories]).T
    if not title:
        title = name
    plt.figure(title)
    plt.plot(tout.T, x, '0.5', lw=2, alpha=0.25) # individual trajectories
    plt.plot(tout[0], x.mean(1), 'k--', lw=3, label="Mean")
    plt.plot(tout[0], x.min(1), 'b--', lw=3, label="Minimum")
    plt.plot(tout[0], x.max(1), 'r--', lw=3, label="Maximum")
    plt.legend(loc=0)
    plt.xlabel('Time')
    plt.ylabel('Population of %s' % name)

tspan = np.linspace(0, 20000, 1000)
sim = StochKitSimulator(model, tspan)
simres = sim.run(n_runs=20, seed=None, algorithm="ssa")

trajectories = simres.all
tout = simres.tout

plot_mean_min_max('Bid_unbound')
plot_mean_min_max('PARP_unbound')
plot_mean_min_max('mSmac_unbound')
plot_mean_min_max('tBid_total')
plot_mean_min_max('CPARP_total')
plot_mean_min_max('cSmac_total')

plt.show()
